# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import numpy as np
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union


def _compute_average_precision(
    recall: np.ndarray,
    precision: np.ndarray,
    use_07_metric: Optional[bool] = False,
) -> float:
    """
    Compute Average Precision (AP) given precision and recall.

    Args:
        recall: Array of recall values.
        precision: Array of precision values.
        use_07_metric: Whether to use the VOC 2007 11-point metric.

    Returns:
        AP score.

    """
    if use_07_metric:
        # 11-point metric
        ap = 0.0
        for t in np.arange(0.0, 1.1, 0.1):
            if np.sum(recall >= t) == 0:
                p = 0
            else:
                p = np.max(precision[recall >= t])
            ap += p / 11.0
    else:
        # correct AP calculation
        # first append sentinel values at the end
        mrec = np.concatenate(([0.0], recall, [1.0]))
        mpre = np.concatenate(([0.0], precision, [0.0]))

        # compute the precision envelope
        for i in range(mpre.size - 1, 0, -1):
            mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

        # to calculate area under PR curve, look for points
        # where X axis (recall) changes value
        i = np.where(mrec[1:] != mrec[:-1])[0]

        # and sum (\Delta recall) * precision
        ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])
    return ap


def _compute_ap_score(
    predicted_bboxes: List[np.ndarray],
    predicted_scores: List[np.ndarray],
    predicted_classes: List[np.ndarray],
    ground_truth_bboxes: List[np.ndarray],
    ground_truth_classes: List[np.ndarray],
    class_id: Optional[int] = None,
    threshold: Optional[float] = 0.5,
) -> float:
    """
    Compute Average Precision (AP) score for a specific class.

    Args:
        predicted_bboxes: List of predicted bounding boxes.
        predicted_scores: List of predicted scores for each bounding box.
        predicted_classes: List of predicted class IDs for each bounding box.
        ground_truth_bboxes: List of ground truth bounding boxes.
        ground_truth_classes: List of true class IDs for each ground truth bounding box.
        class_id: The class ID to compute the AP score.
        threshold: The threshold to determine true positives.

    Returns:
        AP score.

    """
    if class_id is not None:
        ground_truth_bboxes = [
            gt_bbox[gt_class == class_id]
            for gt_class, gt_bbox in zip(ground_truth_classes, ground_truth_bboxes)
        ]
        predicted_bboxes = [
            p_bbox[p_class == class_id]
            for p_class, p_bbox in zip(predicted_classes, predicted_bboxes)
        ]
        predicted_scores = [
            p_score[p_class == class_id]
            for p_class, p_score in zip(predicted_classes, predicted_scores)
        ]
        predicted_indexes = [
            np.array([i] * len(predicted_bboxes[i]))
            for i in range(len(predicted_bboxes))
        ]

    predicted_bboxes = np.concatenate(predicted_bboxes)
    predicted_scores = np.concatenate(predicted_scores)
    predicted_indexes = np.concatenate(predicted_indexes)
    predicted_sort_indexes = np.argsort(-predicted_scores)

    true_positives = np.zeros(predicted_scores.shape[0])
    false_positives = np.zeros(predicted_scores.shape[0])
    gt_bbox_status = defaultdict(set)
    for idx, predicted_sort_index in enumerate(predicted_sort_indexes):
        predicted_index = int(predicted_indexes[predicted_sort_index])
        gt_bboxes = ground_truth_bboxes[predicted_index]
        predicted_bbox = predicted_bboxes[predicted_sort_index]
        max_overlap = -float("inf")
        max_overlap_index = -1
        if gt_bboxes.size > 0:
            ixmin = np.maximum(gt_bboxes[:, 0], predicted_bbox[0])
            iymin = np.maximum(gt_bboxes[:, 1], predicted_bbox[1])
            ixmax = np.minimum(gt_bboxes[:, 2], predicted_bbox[2])
            iymax = np.minimum(gt_bboxes[:, 3], predicted_bbox[3])
            iw = np.maximum(ixmax - ixmin + 1.0, 0.0)
            ih = np.maximum(iymax - iymin + 1.0, 0.0)
            inters = iw * ih
            uni = (
                (predicted_bbox[2] - predicted_bbox[0] + 1.0)
                * (predicted_bbox[3] - predicted_bbox[1] + 1.0)
                + (gt_bboxes[:, 2] - gt_bboxes[:, 0] + 1.0)
                * (gt_bboxes[:, 3] - gt_bboxes[:, 1] + 1.0)
                - inters
            )
            overlaps = inters / uni
            max_overlap = np.max(overlaps)
            max_overlap_index = np.argmax(overlaps)
        if max_overlap > threshold:
            if max_overlap_index not in gt_bbox_status[predicted_index]:
                true_positives[idx] = 1
                gt_bbox_status[predicted_index].add(max_overlap_index)
            else:
                false_positives[idx] = 1
        else:
            false_positives[idx] = 1
    false_positives = np.cumsum(false_positives, axis=0)
    true_positives = np.cumsum(true_positives, axis=0)
    recall = true_positives / float(sum([len(gt) for gt in ground_truth_bboxes]))
    precision = true_positives / np.maximum(
        true_positives + false_positives, np.finfo(np.float64).eps
    )
    ap = _compute_average_precision(recall, precision)
    return ap


def map_score(
    predicted_bboxes: List[np.ndarray],
    predicted_scores: List[np.ndarray],
    predicted_classes: List[np.ndarray],
    ground_truth_bboxes: List[np.ndarray],
    ground_truth_classes: List[np.ndarray],
) -> float:
    """
    Compute Mean Average Precision (mAP) score for all classes.

    Args:
        predicted_bboxes: List of predicted bounding boxes.
        predicted_scores: List of predicted scores for each bounding box.
        predicted_classes: List of predicted class IDs for each bounding box.
        ground_truth_bboxes: List of ground truth bounding boxes.
        ground_truth_classes: List of true class IDs for each ground truth bounding box.

    Returns:
        mAP score.

    """
    unique_classes = set(list(np.concatenate(ground_truth_classes)))
    ap_scores = dict()
    for threshold in range(50, 100, 5):
        ap_scores[threshold] = [
            _compute_ap_score(
                predicted_bboxes,
                predicted_scores,
                predicted_classes,
                ground_truth_bboxes,
                ground_truth_classes,
                class_id,
                threshold / 100,
            )
            for class_id in unique_classes
        ]

    mAP = {iou: np.mean(scores) for iou, scores in ap_scores.items()}
    return np.mean(list(mAP.values()))


def map50_score(
    predicted_bboxes: List[np.ndarray],
    predicted_scores: List[np.ndarray],
    predicted_classes: List[np.ndarray],
    ground_truth_bboxes: List[np.ndarray],
    ground_truth_classes: List[np.ndarray],
) -> float:
    """
    Compute Mean Average Precision (mAP@50) score for all classes.

    Args:
        predicted_bboxes: List of predicted bounding boxes.
        predicted_scores: List of predicted scores for each bounding box.
        predicted_classes: List of predicted class IDs for each bounding box.
        ground_truth_bboxes: List of ground truth bounding boxes.
        ground_truth_classes: List of true class IDs for each ground truth bounding box.

    Returns:
        mAP@50 score.

    """
    unique_classes = set(list(np.concatenate(ground_truth_classes)))
    ap_scores = [
        _compute_ap_score(
            predicted_bboxes,
            predicted_scores,
            predicted_classes,
            ground_truth_bboxes,
            ground_truth_classes,
            class_id,
            0.5,
        )
        for class_id in unique_classes
    ]

    return np.mean(ap_scores)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from torch.cuda.amp import autocast
from unitorch.utils import pop_value, nested_dict_value
from unitorch.models.blip import (
    BlipForPretrain as _BlipForPretrain,
    BlipForClassification as _BlipForClassification,
    BlipForTextClassification as _BlipForTextClassification,
    BlipForImageClassification as _BlipForImageClassification,
    BlipForImageCaption as _BlipForImageCaption,
)
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
    register_model,
)
from unitorch.cli.models import generation_model_decorator
from unitorch.cli.models import ClassificationOutputs, GenerationOutputs, LossOutputs
from unitorch.cli.models.blip import pretrained_blip_infos


@register_model("core/model/pretrain/blip")
class BlipForPretrain(_BlipForPretrain):
    """BLIP model for pretraining."""

    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
        freeze_base_model: Optional[bool] = True,
        gradient_checkpointing: Optional[bool] = False,
        use_all_gather: Optional[bool] = True,
    ):
        """
        Initialize BlipForPretrain.

        Args:
            config_path (str): The path to the model configuration file.
            projection_dim (int, optional): The dimension of the projection head. Defaults to 512.
            freeze_base_model (bool, optional): Whether to freeze the base model parameters. Defaults to True.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
            use_all_gather (bool, optional): Whether to use all_gather operation. Defaults to True.
        """
        super().__init__(
            config_path=config_path,
            projection_dim=projection_dim,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
            use_all_gather=use_all_gather,
        )

    @classmethod
    @add_default_section_for_init("core/model/pretrain/blip")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of BlipForPretrain from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            BlipForPretrain: An instance of BlipForPretrain.
        """
        config.set_default_section("core/model/pretrain/blip")
        pretrained_name = config.getoption("pretrained_name", "default-blip")
        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_blip_infos, pretrained_name, "config"),
        )

        config_path = cached_path(config_path)

        projection_dim = config.getoption("projection_dim", 512)
        freeze_base_model = config.getoption("freeze_base_model", True)
        gradient_checkpointing = config.getoption("gradient_checkpointing", False)
        use_all_gather = config.getoption("use_all_gather", True)

        inst = cls(
            config_path=config_path,
            projection_dim=projection_dim,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
            use_all_gather=use_all_gather,
        )
        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_blip_infos, pretrained_name, "weight"),
            check_none=False,
        )
        if weight_path is not None:
            weight_path = cached_path(weight_path)
            inst.from_pretrained(weight_path)

        return inst

    @autocast()
    def forward(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        attention_mask: torch.Tensor = None,
        position_ids: torch.Tensor = None,
    ):
        """
        Forward pass of the BlipForPretrain model.

        Args:
            input_ids (torch.Tensor): The input token IDs.
            pixel_values (torch.Tensor): The pixel values of the images.
            attention_mask (torch.Tensor, optional): The attention mask. Defaults to None.
            position_ids (torch.Tensor, optional): The position IDs. Defaults to None.

        Returns:
            LossOutputs: The loss outputs.
        """
        outputs = super().forward(
            input_ids=input_ids,
            pixel_values=pixel_values,
            attention_mask=attention_mask,
            position_ids=position_ids,
        )
        return LossOutputs(loss=outputs)


@register_model("core/model/classification/blip")
class BlipForClassification(_BlipForClassification):
    """BLIP model for classification."""

    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
        num_classes: Optional[int] = 1,
        freeze_base_model: Optional[bool] = True,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initialize BlipForClassification.

        Args:
            config_path (str): The path to the model configuration file.
            projection_dim (int, optional): The dimension of the projection head. Defaults to 512.
            num_classes (int, optional): The number of classes for classification. Defaults to 1.
            freeze_base_model (bool, optional): Whether to freeze the base model parameters. Defaults to True.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__(
            config_path=config_path,
            projection_dim=projection_dim,
            num_classes=num_classes,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
        )

    @classmethod
    @add_default_section_for_init("core/model/classification/blip")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of BlipForClassification from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            BlipForClassification: An instance of BlipForClassification.
        """
        config.set_default_section("core/model/classification/blip")
        pretrained_name = config.getoption("pretrained_name", "default-blip")
        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_blip_infos, pretrained_name, "config"),
        )

        config_path = cached_path(config_path)

        projection_dim = config.getoption("projection_dim", 512)
        num_classes = config.getoption("num_classes", 1)
        freeze_base_model = config.getoption("freeze_base_model", True)
        gradient_checkpointing = config.getoption("gradient_checkpointing", False)

        inst = cls(
            config_path=config_path,
            projection_dim=projection_dim,
            num_classes=num_classes,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
        )
        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_blip_infos, pretrained_name, "weight"),
            check_none=False,
        )
        if weight_path is not None:
            weight_path = cached_path(weight_path)
            inst.from_pretrained(weight_path)

        return inst

    @autocast()
    def forward(
        self,
        input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        attention_mask: torch.Tensor = None,
        position_ids: torch.Tensor = None,
    ):
        """
        Forward pass of the BlipForClassification model.

        Args:
            input_ids (torch.Tensor): The input token IDs.
            pixel_values (torch.Tensor): The pixel values of the images.
            attention_mask (torch.Tensor, optional): The attention mask. Defaults to None.
            position_ids (torch.Tensor, optional): The position IDs. Defaults to None.

        Returns:
            ClassificationOutputs: The classification outputs.
        """
        outputs = super().forward(
            input_ids=input_ids,
            pixel_values=pixel_values,
            attention_mask=attention_mask,
            position_ids=position_ids,
        )
        return ClassificationOutputs(outputs=outputs)


@register_model("core/model/classification/blip/text")
class BlipForTextClassification(_BlipForTextClassification):
    """BLIP model for text classification."""

    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
        num_classes: Optional[int] = 1,
        freeze_base_model: Optional[bool] = True,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initialize BlipForTextClassification.

        Args:
            config_path (str): The path to the model configuration file.
            projection_dim (int, optional): The dimension of the projection head. Defaults to 512.
            num_classes (int, optional): The number of classes for classification. Defaults to 1.
            freeze_base_model (bool, optional): Whether to freeze the base model parameters. Defaults to True.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__(
            config_path=config_path,
            projection_dim=projection_dim,
            num_classes=num_classes,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
        )

    @classmethod
    @add_default_section_for_init("core/model/classification/blip/text")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of BlipForTextClassification from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            BlipForTextClassification: An instance of BlipForTextClassification.
        """
        config.set_default_section("core/model/classification/blip/text")
        pretrained_name = config.getoption("pretrained_name", "default-blip")
        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_blip_infos, pretrained_name, "config"),
        )

        config_path = cached_path(config_path)

        projection_dim = config.getoption("projection_dim", 512)
        num_classes = config.getoption("num_classes", 1)
        freeze_base_model = config.getoption("freeze_base_Truemodel", True)
        gradient_checkpointing = config.getoption("gradient_checkpointing", False)

        inst = cls(
            config_path=config_path,
            projection_dim=projection_dim,
            num_classes=num_classes,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
        )
        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_blip_infos, pretrained_name, "weight"),
            check_none=False,
        )
        if weight_path is not None:
            weight_path = cached_path(weight_path)
            inst.from_pretrained(weight_path)

        return inst

    @autocast()
    def forward(
        self,
        input_ids=None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
    ):
        """
        Forward pass of the BlipForTextClassification model.

        Args:
            input_ids (torch.Tensor, optional): The input token IDs. Defaults to None.
            attention_mask (torch.Tensor, optional): The attention mask. Defaults to None.
            position_ids (torch.Tensor, optional): The position IDs. Defaults to None.

        Returns:
            ClassificationOutputs: The classification outputs.
        """
        outputs = super().forward(
            input_ids=input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
        )
        return ClassificationOutputs(outputs=outputs)


@register_model("core/model/classification/blip/image")
class BlipForImageClassification(_BlipForImageClassification):
    """BLIP model for image classification."""

    def __init__(
        self,
        config_path: str,
        projection_dim: Optional[int] = 512,
        num_classes: Optional[int] = 1,
        freeze_base_model: Optional[bool] = True,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initialize BlipForImageClassification.

        Args:
            config_path (str): The path to the model configuration file.
            projection_dim (int, optional): The dimension of the projection head. Defaults to 512.
            num_classes (int, optional): The number of classes for classification. Defaults to 1.
            freeze_base_model (bool, optional): Whether to freeze the base model parameters. Defaults to True.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__(
            config_path=config_path,
            projection_dim=projection_dim,
            num_classes=num_classes,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
        )

    @classmethod
    @add_default_section_for_init("core/model/classification/blip/image")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of BlipForImageClassification from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            BlipForImageClassification: An instance of BlipForImageClassification.
        """
        config.set_default_section("core/model/classification/blip/image")
        pretrained_name = config.getoption("pretrained_name", "default-blip")
        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_blip_infos, pretrained_name, "config"),
        )

        config_path = cached_path(config_path)

        projection_dim = config.getoption("projection_dim", 512)
        num_classes = config.getoption("num_classes", 1)
        freeze_base_model = config.getoption("freeze_base_model", True)
        gradient_checkpointing = config.getoption("gradient_checkpointing", False)

        inst = cls(
            config_path=config_path,
            projection_dim=projection_dim,
            num_classes=num_classes,
            freeze_base_model=freeze_base_model,
            gradient_checkpointing=gradient_checkpointing,
        )
        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_blip_infos, pretrained_name, "weight"),
            check_none=False,
        )
        if weight_path is not None:
            weight_path = cached_path(weight_path)
            inst.from_pretrained(weight_path)

        return inst

    @autocast()
    def forward(
        self,
        pixel_values: torch.Tensor,
    ):
        """
        Forward pass of the BlipForImageClassification model.

        Args:
            pixel_values (torch.Tensor): The pixel values of the images.

        Returns:
            ClassificationOutputs: The classification outputs.
        """
        outputs = super().forward(pixel_values=pixel_values)
        return ClassificationOutputs(outputs=outputs)


@register_model("core/model/caption/blip", generation_model_decorator)
class BlipForImageCaption(_BlipForImageCaption):
    """BLIP model for image captioning."""

    def __init__(
        self,
        config_path: str,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initialize BlipForImageCaption.

        Args:
            config_path (str): The path to the model configuration file.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__(
            config_path=config_path,
            gradient_checkpointing=gradient_checkpointing,
        )

    @classmethod
    @add_default_section_for_init("core/model/caption/blip")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of BlipForImageCaption from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            BlipForImageCaption: An instance of BlipForImageCaption.
        """
        config.set_default_section("core/model/caption/blip")
        pretrained_name = config.getoption("pretrained_name", "default-blip")
        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_blip_infos, pretrained_name, "config"),
        )

        config_path = cached_path(config_path)
        gradient_checkpointing = config.getoption("gradient_checkpointing", False)

        inst = cls(
            config_path=config_path,
            gradient_checkpointing=gradient_checkpointing,
        )
        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_blip_infos, pretrained_name, "weight"),
            check_none=False,
        )
        if weight_path is not None:
            weight_path = cached_path(weight_path)
            inst.from_pretrained(weight_path)

        return inst

    @autocast()
    def forward(
        self,
        pixel_values: torch.Tensor,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
    ):
        """
        Forward pass of the BlipForImageCaption model.

        Args:
            pixel_values (torch.Tensor): The pixel values of the images.
            input_ids (torch.Tensor): The input captions.

        Returns:
            GenerationOutputs: The generation outputs.
        """
        outputs = super().forward(
            pixel_values=pixel_values,
            input_ids=input_ids,
            attention_mask=attention_mask,
        )
        return GenerationOutputs(sequences=outputs)

    @torch.no_grad()
    @autocast()
    def generate(
        self,
        pixel_values: torch.Tensor,
        num_beams: Optional[int] = 5,
        decoder_start_token_id: Optional[int] = 101,
        decoder_end_token_id: Optional[Union[int, List[int]]] = 102,
        num_return_sequences: Optional[int] = 1,
        min_gen_seq_length: Optional[int] = 0,
        max_gen_seq_length: Optional[int] = 48,
        repetition_penalty: Optional[float] = 1.0,
        no_repeat_ngram_size: Optional[int] = 0,
        early_stopping: Optional[bool] = True,
        length_penalty: Optional[float] = 1.0,
        num_beam_groups: Optional[int] = 1,
        diversity_penalty: Optional[float] = 0.0,
        do_sample: Optional[bool] = False,
        temperature: Optional[float] = 1.0,
        top_k: Optional[int] = 50,
        top_p: Optional[float] = 1.0,
    ):
        """
        Generate captions using the BlipForImageCaption model.

        Args:
            pixel_values (torch.Tensor): The pixel values of the images.
            num_beams (int, optional): The number of beams for beam search. Defaults to 5.
            decoder_start_token_id (int, optional): The start token ID for the decoder. Defaults to 30522.
            decoder_end_token_id (int or List[int], optional): The end token ID for the decoder. Defaults to 2.
            num_return_sequences (int, optional): The number of sequences to return. Defaults to 1.
            min_gen_seq_length (int, optional): The minimum length of generated sequences. Defaults to 0.
            max_gen_seq_length (int, optional): The maximum length of generated sequences. Defaults to 48.
            repetition_penalty (float, optional): The repetition penalty. Defaults to 1.0.
            no_repeat_ngram_size (int, optional): The size of n-grams to avoid repeating. Defaults to 0.
            early_stopping (bool, optional): Whether to perform early stopping. Defaults to True.
            length_penalty (float, optional): The length penalty. Defaults to 1.0.
            num_beam_groups (int, optional): The number of beam groups. Defaults to 1.
            diversity_penalty (float, optional): The diversity penalty. Defaults to 0.0.
            do_sample (bool, optional): Whether to use sampling for generation. Defaults to False.
            temperature (float, optional): The temperature value for sampling. Defaults to 1.0.
            top_k (int, optional): The top-k value for sampling. Defaults to 50.
            top_p (float, optional): The top-p value for sampling. Defaults to 1.0.

        Returns:
            GenerationOutputs: The generation outputs.
        """
        outputs = super().generate(
            pixel_values=pixel_values,
            num_beams=num_beams,
            decoder_start_token_id=decoder_start_token_id,
            decoder_end_token_id=decoder_end_token_id,
            num_return_sequences=num_return_sequences,
            min_gen_seq_length=min_gen_seq_length,
            max_gen_seq_length=max_gen_seq_length,
            repetition_penalty=repetition_penalty,
            no_repeat_ngram_size=no_repeat_ngram_size,
            early_stopping=early_stopping,
            length_penalty=length_penalty,
            num_beam_groups=num_beam_groups,
            diversity_penalty=diversity_penalty,
            do_sample=do_sample,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
        )

        return GenerationOutputs(
            sequences=outputs.sequences,
            sequences_scores=outputs.sequences_scores,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.xlm_roberta.modeling import (
    XLMRobertaForClassification,
    XLMRobertaForMaskLM,
    XLMRobertaXLForClassification,
)
from unitorch.models.xlm_roberta.processing import (
    XLMRobertaProcessor,
    get_xlm_roberta_tokenizer,
)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import io
import requests
import time
import base64
import json
import logging
import torch
import torch.nn as nn
import numpy as np
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from random import random

from unitorch.utils import pop_value
from unitorch.cli import (
    add_default_section_for_init,
    add_default_section_for_function,
    register_process,
)
from unitorch.cli import WriterOutputs
from unitorch.cli.models import ACT2FN
from unitorch.cli.models import (
    ModelInputs,
    ModelOutputs,
    ModelTargets,
    TensorsInputs,
    ListTensorsInputs,
    EmbeddingOutputs,
)


def _process_returns(kwargs, dtype="TensorsInputs"):
    assert dtype in globals()
    cls = globals()[dtype]
    assert (
        issubclass(cls, ModelInputs)
        or issubclass(cls, ModelOutputs)
        or issubclass(cls, ModelTargets)
    )
    return cls(**kwargs)


class PreProcessor:
    def __init__(
        self,
    ):
        pass

    @classmethod
    @add_default_section_for_init("core/process")
    def from_core_configure(cls, config, **kwargs):
        pass

    @register_process("core/process/number")
    def _number(
        self,
        text: Union[int, float, str],
        dtype: Optional[str] = "int",
        key: Optional[str] = "num",
        returns: Optional[str] = "TensorsInputs",
    ):
        num = torch.tensor(float(text))
        if dtype == "int":
            num = num.int()
        return _process_returns({key: num}, dtype=returns)

    @register_process("core/process/features")
    def _features(
        self,
        features: Union[List, str],
        sep: Optional[str] = None,
        pad: Optional[float] = None,
        num: Optional[int] = None,
        dtype: Optional[str] = "int",
        shape: Optional[tuple] = None,
        key: Optional[str] = "features",
        returns: Optional[str] = "TensorsInputs",
    ):
        if isinstance(features, str):
            features = features.split(sep=sep)

        features = torch.tensor(list(map(float, features)))
        if dtype == "int":
            features = features.int()

        if pad is not None and num is not None:
            features = features[:num]
            padding = torch.zeros(num - len(features)).to(features) + pad
            features = torch.cat([features, padding])

        if shape is not None:
            features = features.reshape(shape)
        return _process_returns({key: features}, dtype=returns)


class PostProcessor:
    def __init__(
        self,
        act_fn: Optional[str] = None,
    ):
        self.act_fn = ACT2FN.get(act_fn, None)

    @classmethod
    @add_default_section_for_init("core/postprocess")
    def from_core_configure(cls, config, **kwargs):
        pass



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import importlib
from functools import partial
from unitorch.cli import GenericService, registered_service, register_service


# import service modules
import unitorch.cli.services.zip_image



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import torch
import json
import numpy as np
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from transformers import CLIPTokenizer
from torchvision.transforms import (
    Resize,
    CenterCrop,
    RandomCrop,
    ToTensor,
    Normalize,
    Compose,
    RandomHorizontalFlip,
)
from torchvision.transforms.functional import crop
from diffusers.image_processor import VaeImageProcessor
from unitorch.models import HfTextClassificationProcessor, GenericOutputs


class StableXLProcessor:
    def __init__(
        self,
        vocab1_path: str,
        merge1_path: str,
        vocab2_path: str,
        merge2_path: str,
        vae_config_path: Optional[str] = None,
        max_seq_length: Optional[int] = 77,
        position_start_id: Optional[int] = 0,
        pad_token1: Optional[str] = "<|endoftext|>",
        pad_token2: Optional[str] = "!",
        image_size: Optional[int] = 512,
        center_crop: Optional[bool] = False,
    ):
        tokenizer1 = CLIPTokenizer(
            vocab_file=vocab1_path,
            merges_file=merge1_path,
        )

        tokenizer1.cls_token = tokenizer1.bos_token
        tokenizer1.sep_token = tokenizer1.eos_token
        tokenizer1.pad_token = pad_token1

        self.text_processor1 = HfTextClassificationProcessor(
            tokenizer=tokenizer1,
            max_seq_length=max_seq_length,
            position_start_id=position_start_id,
        )

        tokenizer2 = CLIPTokenizer(
            vocab_file=vocab2_path,
            merges_file=merge2_path,
        )

        tokenizer2.cls_token = tokenizer2.bos_token
        tokenizer2.sep_token = tokenizer2.eos_token
        tokenizer2.pad_token = pad_token2

        self.text_processor2 = HfTextClassificationProcessor(
            tokenizer=tokenizer2,
            max_seq_length=max_seq_length,
            position_start_id=position_start_id,
        )

        self.image_size = image_size
        self.center_crop = center_crop
        self.vision_resize = Resize(image_size)
        self.vision_crop = (
            CenterCrop(image_size) if center_crop else RandomCrop(image_size)
        )
        self.vision_flip = RandomHorizontalFlip(p=1.0)
        self.vision_processor = Compose(
            [
                ToTensor(),
                Normalize([0.5], [0.5]),
            ]
        )

        if vae_config_path is not None:
            vae_config_dict = json.load(open(vae_config_path))
            vae_scale_factor = 2 ** (
                len(vae_config_dict.get("block_out_channels", [])) - 1
            )
            self.vae_image_processor = VaeImageProcessor(
                vae_scale_factor=vae_scale_factor
            )

    def text2image(
        self,
        prompt: str,
        image: Union[Image.Image, str],
        prompt2: Optional[str] = None,
        max_seq_length: Optional[int] = None,
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")

        original_size = image.size
        image = self.vision_resize(image)
        if self.center_crop:
            y1 = max(0, int(round((image.height - self.image_size) / 2.0)))
            x1 = max(0, int(round((image.width - self.image_size) / 2.0)))
            image = self.vision_crop(image)
        else:
            y1, x1, h, w = self.vision_crop.get_params(
                image, (self.image_size, self.image_size)
            )
            image = crop(image, y1, x1, h, w)
        if self.vision_flip:
            x1 = image.width - x1
            image = self.vision_flip(image)
        crop_top_left = (y1, x1)
        pixel_values = self.vision_processor(image)

        add_time_ids = (
            original_size + crop_top_left + (self.image_size, self.image_size)
        )

        prompt2 = prompt2 or prompt
        prompt_outputs = self.text_processor1.classification(
            prompt, max_seq_length=max_seq_length
        )
        prompt2_outputs = self.text_processor2.classification(
            prompt2, max_seq_length=max_seq_length
        )

        return GenericOutputs(
            pixel_values=pixel_values,
            input_ids=prompt_outputs.input_ids,
            attention_mask=prompt_outputs.attention_mask,
            input2_ids=prompt2_outputs.input_ids,
            attention2_mask=prompt2_outputs.attention_mask,
            add_time_ids=torch.tensor(add_time_ids),
        )

    def text2image_inputs(
        self,
        prompt: str,
        prompt2: Optional[str] = None,
        negative_prompt: Optional[str] = "",
        negative_prompt2: Optional[str] = None,
        max_seq_length: Optional[int] = None,
    ):
        prompt2 = prompt2 or prompt
        negative_prompt2 = negative_prompt2 or negative_prompt
        prompt_outputs = self.text_processor1.classification(
            prompt, max_seq_length=max_seq_length
        )
        prompt2_outputs = self.text_processor2.classification(
            prompt2, max_seq_length=max_seq_length
        )
        negative_prompt_outputs = self.text_processor1.classification(
            negative_prompt, max_seq_length=max_seq_length
        )
        negative_prompt2_outputs = self.text_processor2.classification(
            negative_prompt2, max_seq_length=max_seq_length
        )
        return GenericOutputs(
            input_ids=prompt_outputs.input_ids,
            attention_mask=prompt_outputs.attention_mask,
            input2_ids=prompt2_outputs.input_ids,
            attention2_mask=prompt2_outputs.attention_mask,
            negative_input_ids=negative_prompt_outputs.input_ids,
            negative_attention_mask=negative_prompt_outputs.attention_mask,
            negative_input2_ids=negative_prompt2_outputs.input_ids,
            negative_attention2_mask=negative_prompt2_outputs.attention_mask,
        )

    def image2image_inputs(
        self,
        prompt: str,
        image: Union[Image.Image, str],
        prompt2: Optional[str] = None,
        negative_prompt: Optional[str] = "",
        negative_prompt2: Optional[str] = None,
        max_seq_length: Optional[int] = None,
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")

        pixel_values = self.vae_image_processor.preprocess(image)[0]

        text_inputs = self.text2image_inputs(
            prompt=prompt,
            prompt2=prompt2,
            negative_prompt=negative_prompt,
            negative_prompt2=negative_prompt2,
            max_seq_length=max_seq_length,
        )

        return GenericOutputs(
            pixel_values=pixel_values,
            **text_inputs,
        )

    def inpainting_inputs(
        self,
        prompt: str,
        image: Union[Image.Image, str],
        mask_image: Union[Image.Image, str],
        prompt2: Optional[str] = None,
        negative_prompt: Optional[str] = "",
        negative_prompt2: Optional[str] = None,
        max_seq_length: Optional[int] = None,
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")

        if isinstance(mask_image, str):
            mask_image = Image.open(mask_image)
        mask_image = mask_image.convert("L")

        pixel_values = self.vae_image_processor.preprocess(image)[0]
        pixel_masks = self.vae_image_processor.preprocess(mask_image)[0]
        pixel_masks = (pixel_masks + 1) / 2

        text_inputs = self.text2image_inputs(
            prompt=prompt,
            prompt2=prompt2,
            negative_prompt=negative_prompt,
            negative_prompt2=negative_prompt2,
            max_seq_length=max_seq_length,
        )

        return GenericOutputs(
            pixel_values=pixel_values,
            pixel_masks=pixel_masks,
            **text_inputs,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from diffusers.schedulers import (
    DPMSolverMultistepScheduler,
    DPMSolverSDEScheduler,
    UniPCMultistepScheduler,
)

Schedulers = {
    "DPM++": DPMSolverMultistepScheduler,
    "DPM++SDE": DPMSolverSDEScheduler,
    "UniPC++": UniPCMultistepScheduler,
}

import unitorch.cli.pipelines.animate
import unitorch.cli.pipelines.blip2
import unitorch.cli.pipelines.blip2controlnet
import unitorch.cli.pipelines.bloom
import unitorch.cli.pipelines.peft
import unitorch.cli.pipelines.stable
import unitorch.cli.pipelines.stable_xl
import unitorch.cli.pipelines.stable_xl_refiner
import unitorch.cli.pipelines.controlnet
import unitorch.cli.pipelines.controlnet_xl
import unitorch.cli.pipelines.llama
import unitorch.cli.pipelines.minigpt4
import unitorch.cli.pipelines.mistral
import unitorch.cli.pipelines.sam



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import numpy as np
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union


def tensor2vid(
    video: torch.Tensor, mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5]
) -> List[np.ndarray]:
    # This code is copied from https://github.com/modelscope/modelscope/blob/1509fdb973e5871f37148a4b5e5964cafd43e64d/modelscope/pipelines/multi_modal/text_to_video_synthesis_pipeline.py#L78
    # reshape to ncfhw
    mean = torch.tensor(mean, device=video.device).reshape(1, -1, 1, 1, 1)
    std = torch.tensor(std, device=video.device).reshape(1, -1, 1, 1, 1)
    # unnormalize back to [0,1]
    video = video.mul_(std).add_(mean)
    video.clamp_(0, 1)
    # prepare the final outputs
    i, c, f, h, w = video.shape
    images = video.permute(2, 3, 0, 4, 1).reshape(
        f, h, i * w, c
    )  # 1st (frames, h, batch_size, w, c) 2nd (frames, h, batch_size * w, c)
    images = images.unbind(dim=0)  # prepare a list of indvidual (consecutive frames)
    images = [
        (image.cpu().numpy() * 255).astype("uint8") for image in images
    ]  # f h w c
    return images



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import io
import torch
import gc
import gradio as gr
from PIL import Image
from unitorch.cli import CoreConfigureParser, GenericWebUI
from unitorch.cli import register_webui
from unitorch.cli.pipelines.minigpt4 import MiniGPT4Blip2LlamaForGenerationPipeline


@register_webui("core/webui/minigpt4")
class MiniGPT4WebUI(GenericWebUI):
    supported_pretrained_names = [
        "minigpt4-7b",
        "minigpt4-13b",
    ]

    def __init__(self, config: CoreConfigureParser):
        self.config = config
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "stopped" if self._pipe is None else "running"
        if len(self.supported_pretrained_names) == 0:
            raise ValueError("No supported pretrained models found.")
        self._name = self.supported_pretrained_names[0]
        self._iface = gr.Blocks()
        with self._iface:
            with gr.Row(variant="panel"):
                pretrained_name = gr.Dropdown(
                    self.supported_pretrained_names,
                    value=self._name,
                    label="Pretrain Checkpoint Name",
                )
                status = gr.Textbox(label="Model Status", value=self._status)
                click_start = gr.Button(value="Start")
                click_stop = gr.Button(value="Stop")
                click_start.click(
                    self.start, inputs=[pretrained_name], outputs=[status]
                )
                click_stop.click(self.stop, outputs=[status])

            with gr.Row(variant="panel"):
                with gr.Column():
                    image = gr.Image(type="pil", label="Input Image")
                    prompt = gr.Textbox(label="Input Prompt")
                    submit = gr.Button(value="Submit")

                caption = gr.Textbox(label="Output Caption")

                submit.click(self.serve, inputs=[prompt, image], outputs=[caption])

            self._iface.load(
                fn=lambda: gr.update(value=self._name), outputs=[pretrained_name]
            )
            self._iface.load(fn=lambda: gr.update(value=self._status), outputs=[status])

    @property
    def name(self):
        return "MiniGPT4"

    @property
    def iface(self):
        return self._iface

    @property
    def status(self):
        return self._status == "running"

    def start(self, pretrained_name, **kwargs):
        if self._status == "running":
            self.stop()
        self.config.set("core/pipeline/minigpt4", "pretrained_name", pretrained_name)
        self._name = pretrained_name
        self._pipe = MiniGPT4Blip2LlamaForGenerationPipeline.from_core_configure(
            self.config
        )
        self._status = "running"
        return self._status

    def stop(self, **kwargs):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "stopped" if self._pipe is None else "running"
        return self._status

    def serve(
        self,
        text: str,
        image: Image.Image,
    ):
        assert self._pipe is not None
        result = self._pipe(text, image)
        return result



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.pipelines.stable_xl_refiner.text2image import (
    StableXLRefinerForText2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.stable_xl_refiner.image2image import (
    StableXLRefinerForImage2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.stable_xl_refiner.inpainting import (
    StableXLRefinerForImageInpaintingPipeline,
)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.utils import pop_value, nested_dict_value
from unitorch.models.diffusers import ControlNetXLProcessor as _ControlNetXLProcessor
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
    register_process,
)
from unitorch.cli.models import (
    TensorsInputs,
)
from unitorch.cli.models.diffusers import pretrained_diffusers_infos


class ControlNetXLProcessor(_ControlNetXLProcessor):
    def __init__(
        self,
        vocab1_path: str,
        merge1_path: str,
        vocab2_path: str,
        merge2_path: str,
        vae_config_path: str,
        max_seq_length: Optional[int] = 77,
        position_start_id: Optional[int] = 0,
        pad_token1: Optional[str] = "<|endoftext|>",
        pad_token2: Optional[str] = "!",
    ):
        super().__init__(
            vocab1_path=vocab1_path,
            merge1_path=merge1_path,
            vocab2_path=vocab2_path,
            merge2_path=merge2_path,
            vae_config_path=vae_config_path,
            max_seq_length=max_seq_length,
            position_start_id=position_start_id,
            pad_token1=pad_token1,
            pad_token2=pad_token2,
        )

    @classmethod
    @add_default_section_for_init("core/process/diffusers/controlnet_xl")
    def from_core_configure(cls, config, **kwargs):
        config.set_default_section("core/process/diffusers/controlnet_xl")
        pretrained_name = config.getoption(
            "pretrained_name", "stable-xl-base-controlnet-canny"
        )
        pretrain_infos = nested_dict_value(pretrained_diffusers_infos, pretrained_name)

        vocab1_path = config.getoption("vocab1_path", None)
        vocab1_path = pop_value(
            vocab1_path,
            nested_dict_value(pretrain_infos, "text", "vocab"),
        )
        vocab1_path = cached_path(vocab1_path)

        merge1_path = config.getoption("merge1_path", None)
        merge1_path = pop_value(
            merge1_path,
            nested_dict_value(pretrain_infos, "text", "merge"),
        )
        merge1_path = cached_path(merge1_path)

        vocab2_path = config.getoption("vocab2_path", None)
        vocab2_path = pop_value(
            vocab2_path,
            nested_dict_value(pretrain_infos, "text2", "vocab"),
        )
        vocab2_path = cached_path(vocab2_path)

        merge2_path = config.getoption("merge2_path", None)
        merge2_path = pop_value(
            merge2_path,
            nested_dict_value(pretrain_infos, "text2", "merge"),
        )
        merge2_path = cached_path(merge2_path)

        vae_config_path = config.getoption("vae_config_path", None)
        vae_config_path = pop_value(
            vae_config_path,
            nested_dict_value(pretrain_infos, "vae", "config"),
        )
        vae_config_path = cached_path(vae_config_path)

        return {
            "vocab1_path": vocab1_path,
            "merge1_path": merge1_path,
            "vocab2_path": vocab2_path,
            "merge2_path": merge2_path,
            "vae_config_path": vae_config_path,
        }

    @register_process("core/process/diffusers/controlnet_xl/text2image")
    def _text2image(
        self,
        prompt: str,
        image: Union[Image.Image, str],
        condition_image: Union[Image.Image, str],
        prompt2: Optional[str] = None,
        max_seq_length: Optional[int] = None,
    ):
        outputs = super().text2image(
            prompt=prompt,
            image=image,
            condition_image=condition_image,
            prompt2=prompt2,
            max_seq_length=max_seq_length,
        )
        return TensorsInputs(
            pixel_values=outputs.pixel_values,
            condition_pixel_values=outputs.condition_pixel_values,
            input_ids=outputs.input_ids,
            attention_mask=outputs.attention_mask,
            input2_ids=outputs.input2_ids,
            attention2_mask=outputs.attention2_mask,
            add_time_ids=outputs.add_time_ids,
        )

    @register_process("core/process/diffusers/controlnet_xl/text2image/inputs")
    def _text2image_inputs(
        self,
        prompt: str,
        condition_image: Union[Image.Image, str],
        negative_prompt: Optional[str] = "",
        max_seq_length: Optional[int] = None,
    ):
        outputs = super().text2image_inputs(
            prompt=prompt,
            condition_image=condition_image,
            negative_prompt=negative_prompt,
            max_seq_length=max_seq_length,
        )
        return TensorsInputs(
            input_ids=outputs.input_ids,
            negative_input_ids=outputs.negative_input_ids,
            condition_pixel_values=outputs.condition_pixel_values,
            attention_mask=outputs.attention_mask,
            negative_attention_mask=outputs.negative_attention_mask,
        )

    @register_process("core/process/diffusers/controlnet_xl/image2image/inputs")
    def _image2image_inputs(
        self,
        prompt: str,
        condition_image: Union[Image.Image, str],
        image: Union[Image.Image, str],
        negative_prompt: Optional[str] = "",
        max_seq_length: Optional[int] = None,
    ):
        outputs = super().image2image_inputs(
            prompt=prompt,
            condition_image=condition_image,
            image=image,
            negative_prompt=negative_prompt,
            max_seq_length=max_seq_length,
        )
        return TensorsInputs(
            input_ids=outputs.input_ids,
            negative_input_ids=outputs.negative_input_ids,
            pixel_values=outputs.pixel_values,
            condition_pixel_values=outputs.condition_pixel_values,
            attention_mask=outputs.attention_mask,
            negative_attention_mask=outputs.negative_attention_mask,
        )

    @register_process("core/process/diffusers/controlnet_xl/inpainting/inputs")
    def _inpainting_inputs(
        prompt: str,
        condition_image: Union[Image.Image, str],
        image: Union[Image.Image, str],
        mask_image: Union[Image.Image, str],
        negative_prompt: Optional[str] = "",
        max_seq_length: Optional[int] = None,
    ):
        outputs = super().inpainting_inputs(
            prompt=prompt,
            condition_image=condition_image,
            image=image,
            mask_image=mask_image,
            negative_prompt=negative_prompt,
            max_seq_length=max_seq_length,
        )
        return TensorsInputs(
            input_ids=outputs.input_ids,
            negative_input_ids=outputs.negative_input_ids,
            pixel_values=outputs.pixel_values,
            condition_pixel_values=outputs.condition_pixel_values,
            attention_mask=outputs.attention_mask,
            negative_attention_mask=outputs.negative_attention_mask,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import re
import fire
import json
import torch
import logging
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from diffusers.models import UNet2DConditionModel
from transformers import CLIPTextConfig, CLIPTextModel
from diffusers.pipelines.stable_diffusion.convert_from_ckpt import (
    convert_controlnet_checkpoint,
    convert_ldm_bert_checkpoint,
    convert_ldm_clip_checkpoint,
    convert_ldm_unet_checkpoint,
    convert_ldm_vae_checkpoint,
    convert_open_clip_checkpoint,
    convert_paint_by_example_checkpoint,
)
from unitorch.cli import cached_path
from unitorch.cli.models.diffusion_utils import load_weight


def stable_clip(
    config_file: str,
    checkpoint_file: str,
    output_checkpoint_file: str,
    base_checkpoint_file: Optional[str] = None,
    is_original_base_checkpoint: bool = False,
):
    def process_name(name: str) -> str:
        name = re.sub(r"lora_te_|lora_up\.|lora_down\.", "", name)
        name = re.sub(
            r"(_model|_attn|_proj|)|_",
            lambda m: "." if m.group(1) is None else m.group(0),
            name,
        )
        return name

    config_file = cached_path(config_file)
    checkpoint_file = cached_path(checkpoint_file)

    config = CLIPTextConfig.from_json_file(config_file)
    text = CLIPTextModel(config)
    text_state_dict = text.state_dict()
    state_dict = {}
    if base_checkpoint_file is not None:
        base_checkpoint_file = cached_path(base_checkpoint_file)
        state_dict = load_weight(base_checkpoint_file)
        if is_original_base_checkpoint:
            state_dict = convert_ldm_clip_checkpoint(
                state_dict, text_encoder=text
            ).state_dict()

    for k, v in state_dict.items():
        if "position_ids" in k:
            continue
        assert k in text_state_dict, f"base checkpoint error, {k} not in text"
        assert (
            text_state_dict[k].shape == v.shape
        ), f"base checkpoint error, {k} shape mismatch, {text_state_dict[k].shape} vs {v.shape}"

    original_state_dict = load_weight(checkpoint_file)
    for k, v in original_state_dict.items():
        if any(x in k for x in ["lora_down", "unet", "conv", "time", "alpha"]):
            continue
        new_k = process_name(k)
        if "lora_up" in k:
            v_up, v_down = v, original_state_dict[k.replace("lora_up", "lora_down")]
            alpha_name = k.replace("lora_up.", "").replace(".weight", ".alpha")
            alpha = original_state_dict[alpha_name].item()
            scale = v_up.shape[1] / alpha
            if v_up.dim() == 2:
                v = torch.matmul(v_up, v_down)
            elif v_up.dim() == 4:
                v = torch.einsum("bchw,cdhw->bdhw", v_up, v_down)
            else:
                assert False, f"dim {v_up.dim()} not supported"
            v = v * scale
        if new_k in text_state_dict:
            assert (
                text_state_dict[new_k].shape == v.shape
            ), f"{new_k} shape mismatch, {text_state_dict[new_k].shape} vs {v.shape}"
            if new_k in state_dict:
                state_dict[new_k] = state_dict[new_k] + v
            else:
                state_dict[new_k] = v
        else:
            logging.warning(f"Skip {new_k} in {checkpoint_file}")


def stable_open_clip(
    config_file: str,
    checkpoint_file: str,
    output_checkpoint_file: str,
    base_checkpoint_file: Optional[str] = None,
    is_original_base_checkpoint: bool = False,
):
    pass


def stable_unet(
    config_file: str,
    checkpoint_file: str,
    output_checkpoint_file: str,
    base_checkpoint_file: Optional[str] = None,
    is_original_base_checkpoint: bool = False,
):
    def process_name(name: str) -> str:
        name = re.sub(r"lora_unet_|lora_up\.|lora_down\.", "", name)
        name = re.sub(
            r"(_block|_k|_v|_q|_in|_out|)|_",
            lambda m: "." if m.group(1) is None else m.group(0),
            name,
        )
        return name

    config_file = cached_path(config_file)
    checkpoint_file = cached_path(checkpoint_file)

    config_dict = json.load(open(config_file))
    unet = UNet2DConditionModel.from_config(config_dict)
    unet_state_dict = unet.state_dict()
    state_dict = {}
    if base_checkpoint_file is not None:
        base_checkpoint_file = cached_path(base_checkpoint_file)
        state_dict = load_weight(base_checkpoint_file)
        if is_original_base_checkpoint:
            has_label_emb = any("label_emb" in k for k in state_dict.keys())
            config_dict["class_embed_type"] = "timestep" if has_label_emb else None
            config_dict["addition_embed_type"] = "text_time" if has_label_emb else None
            state_dict = convert_ldm_unet_checkpoint(state_dict, config=config_dict)

    for k, v in state_dict.items():
        assert k in unet_state_dict, f"base checkpoint error, {k} not in unet"
        assert (
            unet_state_dict[k].shape == v.shape
        ), f"base checkpoint error, {k} shape mismatch, {unet_state_dict[k].shape} vs {v.shape}"

    original_state_dict = load_weight(checkpoint_file)
    for k, v in original_state_dict.items():
        if any(x in k for x in ["lora_down", "text_model", "alpha"]):
            continue
        new_k = process_name(k)
        if "lora_up" in k:
            v_up, v_down = v, original_state_dict[k.replace("lora_up", "lora_down")]
            alpha_name = k.replace("lora_up.", "").replace(".weight", ".alpha")
            alpha = original_state_dict[alpha_name].item()
            scale = v_up.shape[1] / alpha
            if v_up.dim() == 2:
                v = torch.matmul(v_up, v_down)
            elif v_up.dim() == 4:
                v = torch.einsum("bchw,cdhw->bdhw", v_up, v_down)
            else:
                assert False, f"dim {v_up.dim()} not supported"
            v = v * scale
        if new_k in unet_state_dict:
            assert (
                unet_state_dict[new_k].shape == v.shape
            ), f"{new_k} shape mismatch, {unet_state_dict[new_k].shape} vs {v.shape}"
            if new_k in state_dict:
                state_dict[new_k] = state_dict[new_k] + v
            else:
                state_dict[new_k] = v
        else:
            logging.warning(f"Skip {new_k} in {checkpoint_file}")
    torch.save(state_dict, output_checkpoint_file)


def stable_vae(
    config_file: str,
    checkpoint_file: str,
    output_checkpoint_file: str,
    base_checkpoint_file: Optional[str] = None,
    is_original_base_checkpoint: bool = False,
):
    pass


def controlnet(
    config_file: str,
    checkpoint_file: str,
    output_checkpoint_file: str,
    base_checkpoint_file: Optional[str] = None,
    is_original_base_checkpoint: bool = False,
):
    pass


def animate_motion(
    config_file: str,
    checkpoint_file: str,
    output_checkpoint_file: str,
    base_checkpoint_file: str,
):
    base_checkpoint_file = cached_path(base_checkpoint_file)
    state_dict = load_weight(base_checkpoint_file)
    lora_state_dict = load_weight(checkpoint_file)
    for k, v in lora_state_dict.items():
        if "lora.down" in k:
            continue
        up, down = v, lora_state_dict.get(k.replace("_lora.up", "_lora.down"))
        new_k = (
            k.replace("_lora.up", "")
            .replace("unet.", "")
            .replace("processor.", "")
            .replace("to_out", "to_out.0")
        )
        new_v = torch.matmul(up, down)
        state_dict[new_k] += new_v

    torch.save(state_dict, output_checkpoint_file)


if __name__ == "__main__":
    fire.Fire()



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.losses import (
    ListMLELoss as _ListMLELoss,
    ApproxNDCGLoss as _ApproxNDCGLoss,
    ApproxMRRLoss as _ApproxMRRLoss,
)
from unitorch.cli import add_default_section_for_init, register_loss
from unitorch.cli.models import (
    RankingOutputs,
    RankingTargets,
)


@register_loss("core/loss/ranking/listmle")
class ListMLELoss(_ListMLELoss):
    def __init__(
        self,
    ):
        super().__init__()

    @classmethod
    @add_default_section_for_init("core/loss/ranking/listmle")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: RankingOutputs,
        targets: RankingTargets,
    ):
        outputs = outputs.outputs
        masks = targets.masks
        weights = targets.sample_weight
        targets = targets.targets

        return super().forward(
            input=outputs,
            target=targets,
            masks=masks if masks.numel() > 0 else None,
            sample_weight=weights if weights.numel() > 0 else None,
        )


@register_loss("core/loss/ranking/approxndcg")
class ApproxNDCGLoss(_ApproxNDCGLoss):
    def __init__(
        self,
        alpha: Optional[float] = 10.0,
    ):
        super().__init__(
            alpha=alpha,
        )

    @classmethod
    @add_default_section_for_init("core/loss/ranking/approxndcg")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: RankingOutputs,
        targets: RankingTargets,
    ):
        outputs = outputs.outputs
        masks = targets.masks
        weights = targets.sample_weight
        targets = targets.targets

        return super().forward(
            input=outputs,
            target=targets,
            masks=masks if masks.numel() > 0 else None,
            sample_weight=weights if weights.numel() > 0 else None,
        )


@register_loss("core/loss/ranking/approxmrr")
class ApproxMRRLoss(_ApproxMRRLoss):
    def __init__(
        self,
        alpha: Optional[float] = 0.0,
    ):
        super().__init__(
            alpha=alpha,
        )

    @classmethod
    @add_default_section_for_init("core/loss/ranking/approxmrr")
    def from_core_configure(cls, config, **kwargs):
        pass

    def forward(
        self,
        outputs: RankingOutputs,
        targets: RankingTargets,
    ):
        outputs = outputs.outputs
        masks = targets.masks
        weights = targets.sample_weight
        targets = targets.targets

        return super().forward(
            input=outputs,
            target=targets,
            masks=masks if masks.numel() > 0 else None,
            sample_weight=weights if weights.numel() > 0 else None,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import json
import math
import random
import logging
import transformers
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.distributed as dist
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

from transformers.models.blip_2.modeling_blip_2 import (
    Blip2Config,
    Blip2Model,
    Blip2ForConditionalGeneration,
)
from unitorch.utils.decorators import replace
from unitorch.models import GenericModel, GenericOutputs
from unitorch.models.clip.modeling import AllGather



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import torch
import json
import numpy as np
from PIL import Image, ImageFilter
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from transformers import CLIPTokenizer
from torchvision.transforms import (
    Resize,
    CenterCrop,
    ToTensor,
    Normalize,
    Compose,
    RandomHorizontalFlip,
)
from diffusers.image_processor import VaeImageProcessor
from unitorch.models import HfTextClassificationProcessor, GenericOutputs


class MultiControlNetProcessor(HfTextClassificationProcessor):
    def __init__(
        self,
        vocab_path: str,
        merge_path: str,
        vae_config_path: str,
        max_seq_length: Optional[int] = 77,
        position_start_id: Optional[int] = 0,
        pad_token: Optional[str] = "<|endoftext|>",
        image_size: Optional[int] = 512,
    ):
        tokenizer = CLIPTokenizer(
            vocab_file=vocab_path,
            merges_file=merge_path,
        )
        tokenizer.cls_token = tokenizer.bos_token
        tokenizer.sep_token = tokenizer.eos_token
        tokenizer.pad_token = pad_token
        super().__init__(
            tokenizer=tokenizer,
            max_seq_length=max_seq_length,
            position_start_id=position_start_id,
        )
        self.vision_processor = Compose(
            [
                Resize(image_size),
                CenterCrop(image_size),
                ToTensor(),
                Normalize([0.5], [0.5]),
            ]
        )
        self.condition_vision_processor = Compose(
            [
                Resize(image_size),
                CenterCrop(image_size),
                ToTensor(),
            ]
        )

        vae_config_dict = json.load(open(vae_config_path))
        vae_scale_factor = 2 ** (len(vae_config_dict.get("block_out_channels", [])) - 1)
        self.vae_image_processor = VaeImageProcessor(
            vae_scale_factor=vae_scale_factor, do_convert_rgb=True
        )
        self.vae_mask_image_processor = VaeImageProcessor(
            vae_scale_factor=vae_scale_factor
        )
        self.vae_condition_image_processor = VaeImageProcessor(
            vae_scale_factor=vae_scale_factor,
            do_convert_rgb=True,
            do_normalize=False,
        )

    def text2image(
        self,
        image: Union[Image.Image, str],
        condition_images: List[Union[Image.Image, str]],
        prompt: str,
        max_seq_length: Optional[int] = None,
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")
        condition_images = []
        for condition_image in condition_images:
            if isinstance(condition_image, str):
                condition_image = Image.open(condition_image)
            condition_image = condition_image.convert("RGB")
            condition_images.append(condition_image)

        pixel_values = self.vision_processor(image)
        condition_pixel_values = torch.Tensor(
            [
                self.condition_vision_processor(condition_image)
                for condition_image in condition_images
            ]
        )

        prompt_outputs = self.classification(prompt, max_seq_length=max_seq_length)

        return GenericOutputs(
            input_ids=prompt_outputs.input_ids,
            attention_mask=prompt_outputs.attention_mask,
            pixel_values=pixel_values,
            condition_pixel_values=condition_pixel_values,
        )

    def text2image_inputs(
        self,
        condition_images: List[Union[Image.Image, str]],
        prompt: str,
        negative_prompt: Optional[str] = "",
        max_seq_length: Optional[int] = None,
    ):
        condition_images = []
        for condition_image in condition_images:
            if isinstance(condition_image, str):
                condition_image = Image.open(condition_image)
            condition_image = condition_image.convert("RGB")
            condition_images.append(condition_image)

        condition_pixel_values = torch.Tensor(
            [
                self.vae_condition_image_processor.preprocess(condition_image)[0]
                for condition_image in condition_images
            ]
        )

        prompt_outputs = self.classification(prompt, max_seq_length=max_seq_length)
        negative_prompt_outputs = self.classification(
            negative_prompt, max_seq_length=max_seq_length
        )

        return GenericOutputs(
            input_ids=prompt_outputs.input_ids,
            attention_mask=prompt_outputs.attention_mask,
            condition_pixel_values=condition_pixel_values,
            negative_input_ids=negative_prompt_outputs.input_ids,
            negative_attention_mask=negative_prompt_outputs.attention_mask,
        )

    def image2image_inputs(
        self,
        image: Union[Image.Image, str],
        condition_images: List[Union[Image.Image, str]],
        prompt: str,
        negative_prompt: Optional[str] = "",
        max_seq_length: Optional[int] = None,
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")

        pixel_values = self.vae_image_processor.preprocess(image)[0]

        text_inputs = self.text2image_inputs(
            condition_images=condition_images,
            prompt=prompt,
            negative_prompt=negative_prompt,
            max_seq_length=max_seq_length,
        )

        return GenericOutputs(
            pixel_values=pixel_values,
            **text_inputs,
        )

    def inpainting_inputs(
        self,
        image: Union[Image.Image, str],
        mask_image: Union[Image.Image, str],
        condition_images: List[Union[Image.Image, str]],
        prompt: str,
        negative_prompt: Optional[str] = "",
        max_seq_length: Optional[int] = None,
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")

        if isinstance(mask_image, str):
            mask_image = Image.open(mask_image)
        mask_image = mask_image.convert("L")

        pixel_values = self.vae_image_processor.preprocess(image)[0]
        pixel_masks = self.vae_mask_image_processor.preprocess(mask_image)[0]
        pixel_masks = (pixel_masks + 1) / 2

        text_inputs = self.text2image_inputs(
            condition_images=condition_images,
            prompt=prompt,
            negative_prompt=negative_prompt,
            max_seq_length=max_seq_length,
        )
        return GenericOutputs(
            pixel_values=pixel_values,
            pixel_masks=pixel_masks,
            **text_inputs,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import sys
import abc
import logging
import traceback
import importlib
import pkg_resources
import importlib.metadata as importlib_metadata
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from transformers.utils import is_remote_url
from unitorch.utils import cached_path as hf_cached_path
from unitorch.utils import rpartial
from unitorch.cli.core import CoreConfigureParser


def import_library(library):
    is_load_success = False
    try:
        importlib.import_module(library)
        is_load_success = True
    except importlib_metadata.PackageNotFoundError:
        logging.debug(f"import {library} failed.")
        is_load_success = False
    return is_load_success


def cached_path(
    url_or_filename,
    cache_dir: Optional[str] = None,
    force_download: Optional[bool] = False,
    proxies: Optional[str] = None,
    resume_download: Optional[bool] = False,
    user_agent: Union[Dict, str, None] = None,
    extract_compressed_file: Optional[bool] = False,
    force_extract: Optional[bool] = False,
    use_auth_token: Union[bool, str, None] = None,
    local_files_only: Optional[bool] = False,
) -> Optional[str]:
    if not is_remote_url(url_or_filename):
        pkg_filename = pkg_resources.resource_filename("unitorch", url_or_filename)
        if os.path.exists(pkg_filename):
            url_or_filename = pkg_filename

    return hf_cached_path(
        url_or_filename,
        cache_dir=cache_dir,
        force_download=force_download,
        proxies=proxies,
        resume_download=resume_download,
        user_agent=user_agent,
        extract_compressed_file=extract_compressed_file,
        force_extract=force_extract,
        use_auth_token=use_auth_token,
        local_files_only=local_files_only,
    )


# default core config object
global_config = CoreConfigureParser()


def get_global_config():
    return global_config


def set_global_config(config: Union[CoreConfigureParser, str]):
    if isinstance(config, CoreConfigureParser):
        global_config = config
    elif os.path.exists(config):
        global_config = CoreConfigureParser(config)
    else:
        raise ValueError(f"Can't set global config by {config}")


from unitorch.cli.decorators import (
    add_default_section_for_init,
    add_default_section_for_function,
)


# registry function
def registry_func(
    name: str,
    decorators: Union[Callable, List[Callable]] = None,
    save_dict: Optional[Dict] = dict(),
):
    def actual_func(obj):
        save_dict[name] = dict(
            {
                "obj": obj,
                "decorators": decorators,
            }
        )
        return obj

    return actual_func


# register score/dataset/loss/model/optim/writer/scheduler/task
core_modules = [
    "score",
    "dataset",
    "loss",
    "model",
    "optim",
    "writer",
    "scheduler",
    "task",
]

for module in core_modules:
    globals()[f"registered_{module}"] = dict()
    globals()[f"register_{module}"] = partial(
        registry_func,
        save_dict=globals()[f"registered_{module}"],
    )

# register process function
registered_process = dict()


def get_import_module(import_file):
    modules = sys.modules.copy()
    for k, v in modules.items():
        if hasattr(v, "__file__") and v.__file__ == import_file:
            return v
    raise "can't find the module"


def register_process(
    name: str,
    decorators: Union[Callable, List[Callable]] = None,
):
    def actual_func(obj):
        trace_stacks = traceback.extract_stack()
        import_file = trace_stacks[-2][0]
        import_cls_name = trace_stacks[-2][2]
        import_module = get_import_module(import_file)
        registered_process[name] = dict(
            {
                "cls": {
                    "module": import_module,
                    "name": import_cls_name,
                },
                "obj": obj,
                "decorators": decorators,
            }
        )
        return obj

    return actual_func


# init registered modules
def init_registered_module(
    name: str,
    config: CoreConfigureParser,
    registered_module: Dict,
    **kwargs,
):
    if name not in registered_module:
        return

    v = registered_module[name]

    if v["decorators"]:
        return v["decorators"](v["obj"]).from_core_configure(config, **kwargs)
    return v["obj"].from_core_configure(config, **kwargs)


def init_registered_process(
    name: str,
    config: CoreConfigureParser,
    **kwargs,
):
    if name not in registered_process:
        return

    v = registered_process[name]
    cls = getattr(v["cls"]["module"], v["cls"]["name"])
    inst = cls.from_core_configure(config, **kwargs)
    if v["decorators"]:
        return rpartial(v["decorators"](v["obj"]), inst)
    else:
        return rpartial(v["obj"], inst)


# script module
class GenericScript(metaclass=abc.ABCMeta):
    def __init__(self, config: CoreConfigureParser):
        pass

    @abc.abstractmethod
    def launch(self, **kwargs):
        pass


registered_script = dict()
register_script = partial(
    registry_func,
    save_dict=registered_script,
)


# service module
class GenericService(metaclass=abc.ABCMeta):
    def __init__(self, config: CoreConfigureParser):
        pass

    @abc.abstractmethod
    def start(self, **kwargs):
        pass

    @abc.abstractmethod
    def stop(self, **kwargs):
        pass

    @abc.abstractmethod
    def restart(self, **kwargs):
        pass


registered_service = dict()
register_service = partial(
    registry_func,
    save_dict=registered_service,
)


# webui module
class GenericWebUI(metaclass=abc.ABCMeta):
    def __init__(self, config: CoreConfigureParser):
        pass

    @abc.abstractmethod
    def start(self, **kwargs):
        pass

    @abc.abstractmethod
    def stop(self, **kwargs):
        pass


registered_webui = dict()
register_webui = partial(
    registry_func,
    save_dict=registered_webui,
)


# fastapi module
class GenericFastAPI(metaclass=abc.ABCMeta):
    def __init__(self, config: CoreConfigureParser):
        pass

    @abc.abstractproperty
    def router(self):
        pass

    @abc.abstractmethod
    def start(self):
        pass

    @abc.abstractmethod
    def stop(self):
        pass


registered_fastapi = dict()
register_fastapi = partial(
    registry_func,
    save_dict=registered_fastapi,
)


# usful function
from unitorch.cli.writers import WriterMixin, WriterOutputs

# import cli modules
import unitorch.cli.datasets
import unitorch.cli.losses
import unitorch.cli.models
import unitorch.cli.optims
import unitorch.cli.schedulers
import unitorch.cli.scores
import unitorch.cli.tasks
import unitorch.cli.writers



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import json
import torch
import torch.nn.functional as F
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
import safetensors
import diffusers.schedulers as schedulers
from diffusers.models.attention_processor import LoRAAttnProcessor, LoRAAttnProcessor2_0
from transformers import CLIPTextConfig, CLIPTextModel, CLIPTextModelWithProjection
from diffusers.schedulers import SchedulerMixin
from diffusers.models import (
    UNet2DModel,
    UNet2DConditionModel,
    AutoencoderKL,
)
from diffusers.pipelines import (
    DDPMPipeline,
    StableDiffusionXLPipeline,
    StableDiffusionXLImg2ImgPipeline,
    StableDiffusionXLInpaintPipeline,
)
from unitorch.models import (
    GenericModel,
    GenericOutputs,
    QuantizationConfig,
    QuantizationMixin,
)
from unitorch.models.diffusers.modeling_stable import compute_snr


class GenericStableXLRefinerModel(GenericModel, QuantizationMixin):
    prefix_keys_in_state_dict = {
        # unet weights
        "^add_embedding.*": "unet.",
        "^conv_in.*": "unet.",
        "^conv_norm_out.*": "unet.",
        "^conv_out.*": "unet.",
        "^time_embedding.*": "unet.",
        "^up_blocks.*": "unet.",
        "^mid_block.*": "unet.",
        "^down_blocks.*": "unet.",
        # vae weights
        "^encoder.*": "vae.",
        "^decoder.*": "vae.",
        "^post_quant_conv.*": "vae.",
        "^quant_conv.*": "vae.",
    }

    replace_keys_in_state_dict = {
        "\.query\.": ".to_q.",
        "\.key\.": ".to_k.",
        "\.value\.": ".to_v.",
        "\.proj_attn\.": ".to_out.0.",
    }

    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        text2_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        refiner_config_path: Optional[str] = None,
        refiner_text_config_path: Optional[str] = None,
        refiner_text2_config_path: Optional[str] = None,
        refiner_vae_config_path: Optional[str] = None,
        refiner_scheduler_config_path: Optional[str] = None,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        seed: Optional[int] = 1123,
    ):
        super().__init__()
        self.seed = seed
        self.num_train_timesteps = num_train_timesteps
        self.num_infer_timesteps = num_infer_timesteps

        config_dict = json.load(open(config_path))
        if image_size is not None:
            config_dict.update({"sample_size": image_size})
        if in_channels is not None:
            config_dict.update({"in_channels": in_channels})
        if out_channels is not None:
            config_dict.update({"out_channels": out_channels})
        self.unet = UNet2DConditionModel.from_config(config_dict)

        text_config = CLIPTextConfig.from_json_file(text_config_path)
        self.text = CLIPTextModel(text_config)

        text_config2 = CLIPTextConfig.from_json_file(text2_config_path)
        self.text2 = CLIPTextModelWithProjection(text_config2)

        vae_config_dict = json.load(open(vae_config_path))
        self.vae = AutoencoderKL.from_config(vae_config_dict)

        scheduler_config_dict = json.load(open(scheduler_config_path))
        scheduler_class_name = scheduler_config_dict.get("_class_name", "DDPMScheduler")
        assert hasattr(schedulers, scheduler_class_name)
        scheduler_class = getattr(schedulers, scheduler_class_name)
        assert issubclass(scheduler_class, SchedulerMixin)
        scheduler_config_dict["num_train_timesteps"] = num_train_timesteps
        self.scheduler = scheduler_class.from_config(scheduler_config_dict)

        if refiner_config_path is not None:
            refiner_config_dict = json.load(open(refiner_config_path))
            if image_size is not None:
                refiner_config_dict.update({"sample_size": image_size})
            if in_channels is not None:
                refiner_config_dict.update({"in_channels": in_channels})
            if out_channels is not None:
                refiner_config_dict.update({"out_channels": out_channels})
            self.refiner_unet = UNet2DConditionModel.from_config(refiner_config_dict)
        else:
            self.refiner_unet = None

        if refiner_text_config_path is not None:
            refiner_text_config = CLIPTextConfig.from_json_file(
                refiner_text_config_path
            )
            self.refiner_text = CLIPTextModel(refiner_text_config)
        else:
            self.refiner_text = None

        if refiner_text2_config_path is not None:
            refiner_text_config2 = CLIPTextConfig.from_json_file(
                refiner_text2_config_path
            )
            self.refiner_text2 = CLIPTextModelWithProjection(refiner_text_config2)
        else:
            self.refiner_text2 = None

        if refiner_vae_config_path is not None:
            refiner_vae_config_dict = json.load(open(refiner_vae_config_path))
            self.refiner_vae = AutoencoderKL.from_config(refiner_vae_config_dict)
        else:
            self.refiner_vae = None

        if refiner_scheduler_config_path is not None:
            scheduler_config_dict = json.load(open(refiner_scheduler_config_path))
            scheduler_class_name = scheduler_config_dict.get(
                "_class_name", "DDPMScheduler"
            )
            assert hasattr(schedulers, scheduler_class_name)
            scheduler_class = getattr(schedulers, scheduler_class_name)
            assert issubclass(scheduler_class, SchedulerMixin)
            scheduler_config_dict["num_train_timesteps"] = num_train_timesteps
            self.refiner_scheduler = scheduler_class.from_config(scheduler_config_dict)
        else:
            self.refiner_scheduler = None

        if freeze_vae_encoder:
            for param in self.vae.parameters():
                param.requires_grad = False

        if freeze_text_encoder:
            for param in self.text.parameters():
                param.requires_grad = False
            for param in self.text2.parameters():
                param.requires_grad = False

        if quant_config_path is not None:
            self.quant_config = QuantizationConfig.from_json_file(quant_config_path)
            self.quantize(
                self.quant_config,
                ignore_modules=[
                    "lm_head",
                    "unet",
                    "vae",
                    "refiner_unet",
                    "refiner_vae",
                ],
            )

        self.scheduler.set_timesteps(num_inference_steps=self.num_infer_timesteps)
        if self.refiner_scheduler is not None:
            self.refiner_scheduler.set_timesteps(
                num_inference_steps=self.num_infer_timesteps
            )


# base + refiner
class StableXLRefinerForText2ImageGeneration(GenericStableXLRefinerModel):
    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        text2_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        refiner_config_path: Optional[str] = None,
        refiner_text_config_path: Optional[str] = None,
        refiner_text2_config_path: Optional[str] = None,
        refiner_vae_config_path: Optional[str] = None,
        refiner_scheduler_config_path: Optional[str] = None,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        seed: Optional[int] = 1123,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            text2_config_path=text2_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            refiner_config_path=refiner_config_path,
            refiner_text_config_path=refiner_text_config_path,
            refiner_text2_config_path=refiner_text2_config_path,
            refiner_vae_config_path=refiner_vae_config_path,
            refiner_scheduler_config_path=refiner_scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            seed=seed,
        )

        self.pipeline = StableDiffusionXLPipeline(
            vae=self.vae,
            text_encoder=self.text,
            text_encoder_2=self.text2,
            unet=self.unet,
            scheduler=self.scheduler,
            tokenizer=None,
            tokenizer_2=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)

        self.refiner_pipeline = StableDiffusionXLImg2ImgPipeline(
            vae=self.refiner_vae,
            text_encoder=self.refiner_text,
            text_encoder_2=self.refiner_text2,
            unet=self.refiner_unet,
            scheduler=self.refiner_scheduler,
            tokenizer=None,
            tokenizer_2=None,
            requires_aesthetics_score=True,
        )
        self.refiner_pipeline.set_progress_bar_config(disable=True)

    def forward(
        self,
    ):
        raise NotImplementedError

    def generate(
        self,
        input_ids: torch.Tensor,
        input2_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        negative_input2_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        attention2_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        negative_attention2_mask: Optional[torch.Tensor] = None,
        refiner_input_ids: Optional[torch.Tensor] = None,
        refiner_input2_ids: Optional[torch.Tensor] = None,
        refiner_negative_input_ids: Optional[torch.Tensor] = None,
        refiner_negative_input2_ids: Optional[torch.Tensor] = None,
        refiner_attention_mask: Optional[torch.Tensor] = None,
        refiner_attention2_mask: Optional[torch.Tensor] = None,
        refiner_negative_attention_mask: Optional[torch.Tensor] = None,
        refiner_negative_attention2_mask: Optional[torch.Tensor] = None,
        height: Optional[int] = 1024,
        width: Optional[int] = 1024,
        high_noise_frac: Optional[float] = 0.8,
        guidance_scale: Optional[float] = 5.0,
    ):
        prompt_outputs = self.text(
            input_ids,
            # attention_mask,
            output_hidden_states=True,
        )
        prompt_embeds = prompt_outputs.hidden_states[-2]
        negative_prompt_outputs = self.text(
            negative_input_ids,
            # negative_attention_mask,
            output_hidden_states=True,
        )
        negative_prompt_embeds = negative_prompt_outputs.hidden_states[-2]
        prompt2_outputs = self.text2(
            input2_ids,
            # attention2_mask,
            output_hidden_states=True,
        )
        prompt2_embeds = prompt2_outputs.hidden_states[-2]
        negative_prompt2_outputs = self.text2(
            negative_input2_ids,
            # negative_attention2_mask,
            output_hidden_states=True,
        )
        negative_prompt2_embeds = negative_prompt2_outputs.hidden_states[-2]
        prompt_embeds = torch.concat([prompt_embeds, prompt2_embeds], dim=-1)
        negative_prompt_embeds = torch.concat(
            [negative_prompt_embeds, negative_prompt2_embeds], dim=-1
        )
        pooled_prompt_embeds = prompt2_outputs[0]
        negative_pooled_prompt_embeds = negative_prompt2_outputs[0]

        images = self.pipeline(
            prompt_embeds=prompt_embeds,
            negative_prompt_embeds=torch.zeros_like(negative_prompt_embeds).to(
                negative_pooled_prompt_embeds.device
            ),
            pooled_prompt_embeds=pooled_prompt_embeds,
            negative_pooled_prompt_embeds=torch.zeros_like(
                negative_pooled_prompt_embeds
            ).to(negative_pooled_prompt_embeds.device),
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            height=height,
            width=width,
            denoising_end=high_noise_frac,
            guidance_scale=guidance_scale,
            output_type="latent",
        ).images

        if self.refiner_text is not None:
            refiner_prompt_outputs = self.refiner_text(
                refiner_input_ids,
                # refiner_attention_mask,
                output_hidden_states=True,
            )
            refiner_prompt_embeds = refiner_prompt_outputs.hidden_states[-2]
            refiner_negative_prompt_outputs = self.refiner_text(
                refiner_negative_input_ids,
                # negative_attention_mask,
                output_hidden_states=True,
            )
            refiner_negative_prompt_embeds = (
                refiner_negative_prompt_outputs.hidden_states[-2]
            )
            refiner_pooled_prompt_embeds = refiner_prompt_outputs[0]
            refiner_negative_pooled_prompt_embeds = refiner_negative_prompt_outputs[0]

        if self.refiner_text2 is not None:
            refiner_prompt2_outputs = self.refiner_text2(
                refiner_input2_ids,
                # refiner_attention2_mask,
                output_hidden_states=True,
            )
            refiner_prompt2_embeds = refiner_prompt2_outputs.hidden_states[-2]
            refiner_negative_prompt2_outputs = self.refiner_text2(
                refiner_negative_input2_ids,
                # refiner_negative_attention2_mask,
                output_hidden_states=True,
            )
            refiner_negative_prompt2_embeds = (
                refiner_negative_prompt2_outputs.hidden_states[-2]
            )

            refiner_pooled_prompt_embeds = refiner_prompt2_outputs[0]
            refiner_negative_pooled_prompt_embeds = refiner_negative_prompt2_outputs[0]

        if self.refiner_text is not None and self.refiner_text2 is not None:
            refiner_prompt_embeds = torch.concat(
                [refiner_prompt_embeds, refiner_prompt2_embeds], dim=-1
            )
            refiner_negative_prompt_embeds = torch.concat(
                [refiner_negative_prompt_embeds, refiner_negative_prompt2_embeds],
                dim=-1,
            )
        elif self.refiner_text2 is not None:
            refiner_prompt_embeds = refiner_prompt2_embeds
            refiner_negative_prompt_embeds = refiner_negative_prompt2_embeds

        images = self.refiner_pipeline(
            image=images,
            prompt_embeds=refiner_prompt_embeds,
            negative_prompt_embeds=refiner_negative_prompt_embeds,
            pooled_prompt_embeds=refiner_pooled_prompt_embeds,
            negative_pooled_prompt_embeds=refiner_negative_pooled_prompt_embeds,
            generator=torch.Generator(device=self.refiner_pipeline.device).manual_seed(
                self.seed
            ),
            denoising_start=high_noise_frac,
            guidance_scale=guidance_scale,
            target_size=(height, width),
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))


class StableXLRefinerForImage2ImageGeneration(GenericModel, QuantizationMixin):
    pass


class StableXLRefinerForImageInpainting(GenericModel, QuantizationMixin):
    pass



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import re
import torch
import torch.nn as nn
import hashlib
import tempfile
import numpy as np
from PIL import Image
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from diffusers.utils import numpy_to_pil, pt_to_pil, export_to_gif
from unitorch import is_safetensors_available, is_opencv_available
from unitorch.cli import WriterMixin, WriterOutputs
from unitorch.cli import (
    add_default_section_for_init,
    add_default_section_for_function,
    register_process,
)
from unitorch.cli.models.modeling_utils import TensorsOutputs, TensorsTargets

from unitorch.cli import cached_path


def load_weight(
    path,
    replace_keys: Optional[Dict] = dict(),
    prefix_keys: Optional[Dict] = dict(),
):
    if path.endswith(".safetensors"):
        assert is_safetensors_available(), "Please install safetensors first."
        import safetensors

        path = cached_path(path)
        state_dict = safetensors.torch.load_file(path)
    else:
        path = cached_path(path)
        state_dict = torch.load(path, map_location="cpu")

    results = dict()
    for key, value in list(state_dict.items()):
        for rkey, prefix in prefix_keys.items():
            if re.match(rkey, key):
                key = prefix + key
                break

        for rkey, nkey in replace_keys.items():
            key = re.sub(rkey, nkey, key)

        results[key] = value

    return results


def numpy2vid(video, mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5]):
    batch_size, channels, num_frames, height, width = video.shape
    mean = np.array(mean).reshape(1, -1, 1, 1, 1)
    std = np.array(std).reshape(1, -1, 1, 1, 1)
    # unnormalize back to [0,1]
    video = video * std + mean
    video = np.clip(video, 0, 1)
    # prepare the final outputs
    i, c, f, h, w = video.shape
    images = np.transpose(video, (2, 3, 0, 4, 1)).reshape(f, h, i * w, c)
    images = np.split(images, f, axis=0)
    images = [(image.squeeze(0) * 255).astype("uint8") for image in images]  # f h w c
    return images


def export_to_video(
    video_frames: List[np.ndarray], output_video_path: str = None
) -> str:
    assert is_opencv_available(), "Please install python3-opencv first."
    import cv2

    if output_video_path is None:
        output_video_path = tempfile.NamedTemporaryFile(suffix=".mp4").name

    fourcc = cv2.VideoWriter_fourcc(*"avc1")
    h, w, c = video_frames[0].shape
    video_writer = cv2.VideoWriter(output_video_path, fourcc, fps=8, frameSize=(w, h))
    for i in range(len(video_frames)):
        img = cv2.cvtColor(video_frames[i], cv2.COLOR_RGB2BGR)
        video_writer.write(img)
    return output_video_path


@dataclass
class DiffusionOutputs(TensorsOutputs, WriterMixin):
    outputs: torch.Tensor


@dataclass
class DiffusionTargets(TensorsTargets):
    targets: torch.Tensor
    masks: Optional[torch.Tensor] = torch.empty(0)


class DiffusionProcessor:
    def __init__(self, output_folder: Optional[str] = None):
        assert output_folder is not None
        self.output_folder = output_folder
        if not os.path.exists(output_folder):
            os.makedirs(self.output_folder, exist_ok=True)

    @classmethod
    @add_default_section_for_init("core/process/diffusers")
    def from_core_configure(cls, config, **kwargs):
        pass

    def save_image(self, image: Image.Image):
        md5 = hashlib.md5()
        md5.update(image.tobytes())
        name = md5.hexdigest() + ".jpg"
        output_path = f"{self.output_folder}/{name}"
        image.save(output_path)
        return name

    def save_gif(self, images: List[Image.Image]):
        md5 = hashlib.md5()
        for image in images:
            md5.update(image.tobytes())
        name = md5.hexdigest() + ".gif"
        output_path = f"{self.output_folder}/{name}"
        export_to_gif(images, output_path)
        return name

    def save_video(self, video):
        md5 = hashlib.md5()
        for image in video:
            md5.update(image.tobytes())
        name = md5.hexdigest() + ".mp4"
        output_path = f"{self.output_folder}/{name}"
        export_to_video(video, output_path)
        return name

    @register_process("core/postprocess/diffusion/image")
    def _diffusion_image(
        self,
        outputs: DiffusionOutputs,
    ):
        results = outputs.to_pandas()
        assert results.shape[0] == 0 or results.shape[0] == outputs.outputs.shape[0]
        images = outputs.outputs.numpy()
        images = numpy_to_pil(images)
        results["diffusion"] = [self.save_image(image) for image in images]
        return WriterOutputs(results)

    @register_process("core/postprocess/diffusion/gif")
    def _diffusion_gif(
        self,
        outputs: DiffusionOutputs,
    ):
        results = outputs.to_pandas()
        assert results.shape[0] == 0 or results.shape[0] == outputs.outputs.shape[0]
        images = outputs.outputs.numpy()
        images = [numpy_to_pil(image) for image in images]
        results["diffusion"] = [self.save_gif(image) for image in images]
        return WriterOutputs(results)

    @register_process("core/postprocess/diffusion/video")
    def _diffusion_video(
        self,
        outputs: DiffusionOutputs,
    ):
        results = outputs.to_pandas()
        assert results.shape[0] == 0 or results.shape[0] == outputs.outputs.shape[0]
        videos = outputs.outputs.numpy()
        videos = [numpy2vid(video) for video in videos]
        results["diffusion"] = [self.save_video(video) for video in videos]
        return WriterOutputs(results)


def diffusion_model_decorator(cls):
    class DiffusionModel(nn.Module):
        def __init__(self, *args, **kwargs):
            super().__init__()
            if "__diffusion_model__" in kwargs:
                self.model = kwargs.pop("__diffusion_model__")
            else:
                self.model = cls(*args, **kwargs)

            __more_attrs__ = [
                "load_state_dict",
                "state_dict",
                "save_checkpoint",
                "from_checkpoint",
                "from_pretrained",
            ]
            for __more_attr__ in __more_attrs__:
                setattr(self, __more_attr__, getattr(self.model, __more_attr__))

            assert hasattr(self.model, "generate")

            self.model.register_forward_hook(self._hook)
            self.__in_training__ = False

        def _hook(self, module, inputs, outputs):
            self.__in_training__ = True

        def forward(self, *args, **kwargs):
            if self.training or self.__in_training__:
                return self.model(*args, **kwargs)
            return self.model.generate(*args, **kwargs)

        @classmethod
        def from_core_configure(_cls, cfg, **kwargs):
            model = cls.from_core_configure(cfg, **kwargs)
            return _cls(__diffusion_model__=model)

    return DiffusionModel



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
from unitorch.models import GenericModel


class REBNCONV(nn.Module):
    def __init__(self, in_ch=3, out_ch=3, dirate=1, stride=1):
        super().__init__()

        self.conv_s1 = nn.Conv2d(
            in_ch, out_ch, 3, padding=1 * dirate, dilation=1 * dirate, stride=stride
        )
        self.bn_s1 = nn.BatchNorm2d(out_ch)
        self.relu_s1 = nn.ReLU(inplace=True)

    def forward(self, x):
        hx = x
        xout = self.relu_s1(self.bn_s1(self.conv_s1(hx)))

        return xout


## upsample tensor 'src' to have the same spatial size with tensor 'tar'
def _upsample_like(src, tar):
    src = F.interpolate(src, size=tar.shape[2:], mode="bilinear")

    return src


### RSU-7 ###
class RSU7(nn.Module):
    def __init__(self, in_ch=3, mid_ch=12, out_ch=3, img_size=512):
        super().__init__()

        self.in_ch = in_ch
        self.mid_ch = mid_ch
        self.out_ch = out_ch

        self.rebnconvin = REBNCONV(in_ch, out_ch, dirate=1)  ## 1 -> 1/2

        self.rebnconv1 = REBNCONV(out_ch, mid_ch, dirate=1)
        self.pool1 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.rebnconv2 = REBNCONV(mid_ch, mid_ch, dirate=1)
        self.pool2 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.rebnconv3 = REBNCONV(mid_ch, mid_ch, dirate=1)
        self.pool3 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.rebnconv4 = REBNCONV(mid_ch, mid_ch, dirate=1)
        self.pool4 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.rebnconv5 = REBNCONV(mid_ch, mid_ch, dirate=1)
        self.pool5 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.rebnconv6 = REBNCONV(mid_ch, mid_ch, dirate=1)

        self.rebnconv7 = REBNCONV(mid_ch, mid_ch, dirate=2)

        self.rebnconv6d = REBNCONV(mid_ch * 2, mid_ch, dirate=1)
        self.rebnconv5d = REBNCONV(mid_ch * 2, mid_ch, dirate=1)
        self.rebnconv4d = REBNCONV(mid_ch * 2, mid_ch, dirate=1)
        self.rebnconv3d = REBNCONV(mid_ch * 2, mid_ch, dirate=1)
        self.rebnconv2d = REBNCONV(mid_ch * 2, mid_ch, dirate=1)
        self.rebnconv1d = REBNCONV(mid_ch * 2, out_ch, dirate=1)

    def forward(self, x):
        b, c, h, w = x.shape

        hx = x
        hxin = self.rebnconvin(hx)

        hx1 = self.rebnconv1(hxin)
        hx = self.pool1(hx1)

        hx2 = self.rebnconv2(hx)
        hx = self.pool2(hx2)

        hx3 = self.rebnconv3(hx)
        hx = self.pool3(hx3)

        hx4 = self.rebnconv4(hx)
        hx = self.pool4(hx4)

        hx5 = self.rebnconv5(hx)
        hx = self.pool5(hx5)

        hx6 = self.rebnconv6(hx)

        hx7 = self.rebnconv7(hx6)

        hx6d = self.rebnconv6d(torch.cat((hx7, hx6), 1))
        hx6dup = _upsample_like(hx6d, hx5)

        hx5d = self.rebnconv5d(torch.cat((hx6dup, hx5), 1))
        hx5dup = _upsample_like(hx5d, hx4)

        hx4d = self.rebnconv4d(torch.cat((hx5dup, hx4), 1))
        hx4dup = _upsample_like(hx4d, hx3)

        hx3d = self.rebnconv3d(torch.cat((hx4dup, hx3), 1))
        hx3dup = _upsample_like(hx3d, hx2)

        hx2d = self.rebnconv2d(torch.cat((hx3dup, hx2), 1))
        hx2dup = _upsample_like(hx2d, hx1)

        hx1d = self.rebnconv1d(torch.cat((hx2dup, hx1), 1))

        return hx1d + hxin


### RSU-6 ###
class RSU6(nn.Module):
    def __init__(self, in_ch=3, mid_ch=12, out_ch=3):
        super().__init__()

        self.rebnconvin = REBNCONV(in_ch, out_ch, dirate=1)

        self.rebnconv1 = REBNCONV(out_ch, mid_ch, dirate=1)
        self.pool1 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.rebnconv2 = REBNCONV(mid_ch, mid_ch, dirate=1)
        self.pool2 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.rebnconv3 = REBNCONV(mid_ch, mid_ch, dirate=1)
        self.pool3 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.rebnconv4 = REBNCONV(mid_ch, mid_ch, dirate=1)
        self.pool4 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.rebnconv5 = REBNCONV(mid_ch, mid_ch, dirate=1)

        self.rebnconv6 = REBNCONV(mid_ch, mid_ch, dirate=2)

        self.rebnconv5d = REBNCONV(mid_ch * 2, mid_ch, dirate=1)
        self.rebnconv4d = REBNCONV(mid_ch * 2, mid_ch, dirate=1)
        self.rebnconv3d = REBNCONV(mid_ch * 2, mid_ch, dirate=1)
        self.rebnconv2d = REBNCONV(mid_ch * 2, mid_ch, dirate=1)
        self.rebnconv1d = REBNCONV(mid_ch * 2, out_ch, dirate=1)

    def forward(self, x):
        hx = x

        hxin = self.rebnconvin(hx)

        hx1 = self.rebnconv1(hxin)
        hx = self.pool1(hx1)

        hx2 = self.rebnconv2(hx)
        hx = self.pool2(hx2)

        hx3 = self.rebnconv3(hx)
        hx = self.pool3(hx3)

        hx4 = self.rebnconv4(hx)
        hx = self.pool4(hx4)

        hx5 = self.rebnconv5(hx)

        hx6 = self.rebnconv6(hx5)

        hx5d = self.rebnconv5d(torch.cat((hx6, hx5), 1))
        hx5dup = _upsample_like(hx5d, hx4)

        hx4d = self.rebnconv4d(torch.cat((hx5dup, hx4), 1))
        hx4dup = _upsample_like(hx4d, hx3)

        hx3d = self.rebnconv3d(torch.cat((hx4dup, hx3), 1))
        hx3dup = _upsample_like(hx3d, hx2)

        hx2d = self.rebnconv2d(torch.cat((hx3dup, hx2), 1))
        hx2dup = _upsample_like(hx2d, hx1)

        hx1d = self.rebnconv1d(torch.cat((hx2dup, hx1), 1))

        return hx1d + hxin


### RSU-5 ###
class RSU5(nn.Module):
    def __init__(self, in_ch=3, mid_ch=12, out_ch=3):
        super().__init__()

        self.rebnconvin = REBNCONV(in_ch, out_ch, dirate=1)

        self.rebnconv1 = REBNCONV(out_ch, mid_ch, dirate=1)
        self.pool1 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.rebnconv2 = REBNCONV(mid_ch, mid_ch, dirate=1)
        self.pool2 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.rebnconv3 = REBNCONV(mid_ch, mid_ch, dirate=1)
        self.pool3 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.rebnconv4 = REBNCONV(mid_ch, mid_ch, dirate=1)

        self.rebnconv5 = REBNCONV(mid_ch, mid_ch, dirate=2)

        self.rebnconv4d = REBNCONV(mid_ch * 2, mid_ch, dirate=1)
        self.rebnconv3d = REBNCONV(mid_ch * 2, mid_ch, dirate=1)
        self.rebnconv2d = REBNCONV(mid_ch * 2, mid_ch, dirate=1)
        self.rebnconv1d = REBNCONV(mid_ch * 2, out_ch, dirate=1)

    def forward(self, x):
        hx = x

        hxin = self.rebnconvin(hx)

        hx1 = self.rebnconv1(hxin)
        hx = self.pool1(hx1)

        hx2 = self.rebnconv2(hx)
        hx = self.pool2(hx2)

        hx3 = self.rebnconv3(hx)
        hx = self.pool3(hx3)

        hx4 = self.rebnconv4(hx)

        hx5 = self.rebnconv5(hx4)

        hx4d = self.rebnconv4d(torch.cat((hx5, hx4), 1))
        hx4dup = _upsample_like(hx4d, hx3)

        hx3d = self.rebnconv3d(torch.cat((hx4dup, hx3), 1))
        hx3dup = _upsample_like(hx3d, hx2)

        hx2d = self.rebnconv2d(torch.cat((hx3dup, hx2), 1))
        hx2dup = _upsample_like(hx2d, hx1)

        hx1d = self.rebnconv1d(torch.cat((hx2dup, hx1), 1))

        return hx1d + hxin


### RSU-4 ###
class RSU4(nn.Module):
    def __init__(self, in_ch=3, mid_ch=12, out_ch=3):
        super().__init__()

        self.rebnconvin = REBNCONV(in_ch, out_ch, dirate=1)

        self.rebnconv1 = REBNCONV(out_ch, mid_ch, dirate=1)
        self.pool1 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.rebnconv2 = REBNCONV(mid_ch, mid_ch, dirate=1)
        self.pool2 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.rebnconv3 = REBNCONV(mid_ch, mid_ch, dirate=1)

        self.rebnconv4 = REBNCONV(mid_ch, mid_ch, dirate=2)

        self.rebnconv3d = REBNCONV(mid_ch * 2, mid_ch, dirate=1)
        self.rebnconv2d = REBNCONV(mid_ch * 2, mid_ch, dirate=1)
        self.rebnconv1d = REBNCONV(mid_ch * 2, out_ch, dirate=1)

    def forward(self, x):
        hx = x

        hxin = self.rebnconvin(hx)

        hx1 = self.rebnconv1(hxin)
        hx = self.pool1(hx1)

        hx2 = self.rebnconv2(hx)
        hx = self.pool2(hx2)

        hx3 = self.rebnconv3(hx)

        hx4 = self.rebnconv4(hx3)

        hx3d = self.rebnconv3d(torch.cat((hx4, hx3), 1))
        hx3dup = _upsample_like(hx3d, hx2)

        hx2d = self.rebnconv2d(torch.cat((hx3dup, hx2), 1))
        hx2dup = _upsample_like(hx2d, hx1)

        hx1d = self.rebnconv1d(torch.cat((hx2dup, hx1), 1))

        return hx1d + hxin


### RSU-4F ###
class RSU4F(nn.Module):
    def __init__(self, in_ch=3, mid_ch=12, out_ch=3):
        super().__init__()

        self.rebnconvin = REBNCONV(in_ch, out_ch, dirate=1)

        self.rebnconv1 = REBNCONV(out_ch, mid_ch, dirate=1)
        self.rebnconv2 = REBNCONV(mid_ch, mid_ch, dirate=2)
        self.rebnconv3 = REBNCONV(mid_ch, mid_ch, dirate=4)

        self.rebnconv4 = REBNCONV(mid_ch, mid_ch, dirate=8)

        self.rebnconv3d = REBNCONV(mid_ch * 2, mid_ch, dirate=4)
        self.rebnconv2d = REBNCONV(mid_ch * 2, mid_ch, dirate=2)
        self.rebnconv1d = REBNCONV(mid_ch * 2, out_ch, dirate=1)

    def forward(self, x):
        hx = x

        hxin = self.rebnconvin(hx)

        hx1 = self.rebnconv1(hxin)
        hx2 = self.rebnconv2(hx1)
        hx3 = self.rebnconv3(hx2)

        hx4 = self.rebnconv4(hx3)

        hx3d = self.rebnconv3d(torch.cat((hx4, hx3), 1))
        hx2d = self.rebnconv2d(torch.cat((hx3d, hx2), 1))
        hx1d = self.rebnconv1d(torch.cat((hx2d, hx1), 1))

        return hx1d + hxin


class myrebnconv(nn.Module):
    def __init__(
        self,
        in_ch=3,
        out_ch=1,
        kernel_size=3,
        stride=1,
        padding=1,
        dilation=1,
        groups=1,
    ):
        super().__init__()

        self.conv = nn.Conv2d(
            in_ch,
            out_ch,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
        )
        self.bn = nn.BatchNorm2d(out_ch)
        self.rl = nn.ReLU(inplace=True)

    def forward(self, x):
        return self.rl(self.bn(self.conv(x)))


class BRIAForSegmentation(GenericModel):
    def __init__(
        self,
        in_ch: int = 3,
        out_ch: int = 1,
    ):
        super().__init__()
        self.conv_in = nn.Conv2d(in_ch, 64, 3, stride=2, padding=1)
        self.pool_in = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.stage1 = RSU7(64, 32, 64)
        self.pool12 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.stage2 = RSU6(64, 32, 128)
        self.pool23 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.stage3 = RSU5(128, 64, 256)
        self.pool34 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.stage4 = RSU4(256, 128, 512)
        self.pool45 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.stage5 = RSU4F(512, 256, 512)
        self.pool56 = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        self.stage6 = RSU4F(512, 256, 512)

        # decoder
        self.stage5d = RSU4F(1024, 256, 512)
        self.stage4d = RSU4(1024, 128, 256)
        self.stage3d = RSU5(512, 64, 128)
        self.stage2d = RSU6(256, 32, 64)
        self.stage1d = RSU7(128, 16, 64)

        self.side1 = nn.Conv2d(64, out_ch, 3, padding=1)
        # self.side2 = nn.Conv2d(64, out_ch, 3, padding=1)
        # self.side3 = nn.Conv2d(128, out_ch, 3, padding=1)
        # self.side4 = nn.Conv2d(256, out_ch, 3, padding=1)
        # self.side5 = nn.Conv2d(512, out_ch, 3, padding=1)
        # self.side6 = nn.Conv2d(512, out_ch, 3, padding=1)

        # self.outconv = nn.Conv2d(6*out_ch,out_ch,1)

    def forward(self, x):
        hx = x

        hxin = self.conv_in(hx)
        # hx = self.pool_in(hxin)

        # stage 1
        hx1 = self.stage1(hxin)
        hx = self.pool12(hx1)

        # stage 2
        hx2 = self.stage2(hx)
        hx = self.pool23(hx2)

        # stage 3
        hx3 = self.stage3(hx)
        hx = self.pool34(hx3)

        # stage 4
        hx4 = self.stage4(hx)
        hx = self.pool45(hx4)

        # stage 5
        hx5 = self.stage5(hx)
        hx = self.pool56(hx5)

        # stage 6
        hx6 = self.stage6(hx)
        hx6up = _upsample_like(hx6, hx5)

        # -------------------- decoder --------------------
        hx5d = self.stage5d(torch.cat((hx6up, hx5), 1))
        hx5dup = _upsample_like(hx5d, hx4)

        hx4d = self.stage4d(torch.cat((hx5dup, hx4), 1))
        hx4dup = _upsample_like(hx4d, hx3)

        hx3d = self.stage3d(torch.cat((hx4dup, hx3), 1))
        hx3dup = _upsample_like(hx3d, hx2)

        hx2d = self.stage2d(torch.cat((hx3dup, hx2), 1))
        hx2dup = _upsample_like(hx2d, hx1)

        hx1d = self.stage1d(torch.cat((hx2dup, hx1), 1))

        # side output
        d1 = self.side1(hx1d)
        d1 = _upsample_like(d1, x)

        # d2 = self.side2(hx2d)
        # d2 = _upsample_like(d2, x)

        # d3 = self.side3(hx3d)
        # d3 = _upsample_like(d3, x)

        # d4 = self.side4(hx4d)
        # d4 = _upsample_like(d4, x)

        # d5 = self.side5(hx5d)
        # d5 = _upsample_like(d5, x)

        # d6 = self.side6(hx6)
        # d6 = _upsample_like(d6, x)

        # return [
        #     F.sigmoid(d1),
        #     F.sigmoid(d2),
        #     F.sigmoid(d3),
        #     F.sigmoid(d4),
        #     F.sigmoid(d5),
        #     F.sigmoid(d6),
        # ], [hx1d, hx2d, hx3d, hx4d, hx5d, hx6]
        return F.sigmoid(d1)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import torch
from transformers import MBartTokenizer

from unitorch.utils import pop_value
from unitorch.models import HfTextGenerationProcessor


def get_mbart_tokenizer(
    vocab_path,
    special_input_ids=dict(),
):
    """
    Returns an MBartTokenizer initialized with the provided vocabulary file and special input IDs.

    Args:
        vocab_path (str): The path to the vocabulary file.
        special_input_ids (Dict, optional): A dictionary of special input IDs. Defaults to an empty dictionary.

    Returns:
        MBartTokenizer: The initialized MBartTokenizer.
    """
    assert os.path.exists(vocab_path)
    tokenizer = MBartTokenizer(vocab_path)
    for token, _id in special_input_ids.items():
        tokenizer.added_tokens_encoder[token] = _id
        tokenizer.unique_no_split_tokens.append(token)
        tokenizer.added_tokens_decoder[_id] = token
        tokenizer.add_tokens(token, special_tokens=True)
    return tokenizer


class MBartProcessor(HfTextGenerationProcessor):
    def __init__(
        self,
        vocab_path: str,
        max_seq_length: Optional[int] = 128,
        max_gen_seq_length: Optional[int] = 48,
        special_input_ids: Optional[Dict] = dict(),
    ):
        """
        Initializes an MBartProcessor with the provided parameters.

        Args:
            vocab_path (str): The path to the vocabulary file.
            max_seq_length (int, optional): The maximum sequence length. Defaults to 128.
            max_gen_seq_length (int, optional): The maximum generation sequence length. Defaults to 48.
            special_input_ids (Dict, optional): A dictionary of special input IDs. Defaults to an empty dictionary.
        """
        tokenizer = get_mbart_tokenizer(
            vocab_path,
            special_input_ids=special_input_ids,
        )
        super().__init__(
            tokenizer=tokenizer,
            max_seq_length=max_seq_length,
            max_gen_seq_length=max_gen_seq_length,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

# pretrained infos
pretrained_mbart_infos = {
    "default-mbart": {
        "config": "https://huggingface.co/facebook/mbart-large-cc25/resolve/main/config.json",
        "vocab": "https://huggingface.co/facebook/mbart-large-cc25/resolve/main/sentence.bpe.model",
    },
    "mbart-large-cc25": {
        "config": "https://huggingface.co/facebook/mbart-large-cc25/resolve/main/config.json",
        "vocab": "https://huggingface.co/facebook/mbart-large-cc25/resolve/main/sentence.bpe.model",
        "weight": "https://huggingface.co/facebook/mbart-large-cc25/resolve/main/pytorch_model.bin",
    },
}

import unitorch.cli.models.mbart.modeling
import unitorch.cli.models.mbart.processing
from unitorch.cli.models.mbart.modeling import MBartForGeneration
from unitorch.cli.models.mbart.processing import MBartProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.pipelines.controlnet_xl.text2image import (
    ControlNetXLForText2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.controlnet_xl.image2image import (
    ControlNetXLForImage2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.controlnet_xl.inpainting import (
    ControlNetXLForImageInpaintingPipeline,
)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import torch
import json
import numpy as np
from PIL import Image, ImageFilter
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from transformers import CLIPTokenizer
from torchvision.transforms.functional import crop
from torchvision.transforms import (
    Resize,
    CenterCrop,
    ToTensor,
    Normalize,
    Compose,
    RandomHorizontalFlip,
)
from diffusers.image_processor import VaeImageProcessor
from unitorch.models import HfTextClassificationProcessor, GenericOutputs


class ControlNetXLProcessor:
    def __init__(
        self,
        vocab1_path: str,
        merge1_path: str,
        vocab2_path: str,
        merge2_path: str,
        vae_config_path: str,
        max_seq_length: Optional[int] = 77,
        position_start_id: Optional[int] = 0,
        pad_token1: Optional[str] = "<|endoftext|>",
        pad_token2: Optional[str] = "!",
        image_size: Optional[int] = 512,
    ):
        tokenizer1 = CLIPTokenizer(
            vocab_file=vocab1_path,
            merges_file=merge1_path,
        )

        tokenizer1.cls_token = tokenizer1.bos_token
        tokenizer1.sep_token = tokenizer1.eos_token
        tokenizer1.pad_token = pad_token1

        self.text_processor1 = HfTextClassificationProcessor(
            tokenizer=tokenizer1,
            max_seq_length=max_seq_length,
            position_start_id=position_start_id,
        )

        tokenizer2 = CLIPTokenizer(
            vocab_file=vocab2_path,
            merges_file=merge2_path,
        )

        tokenizer2.cls_token = tokenizer2.bos_token
        tokenizer2.sep_token = tokenizer2.eos_token
        tokenizer2.pad_token = pad_token2

        self.text_processor2 = HfTextClassificationProcessor(
            tokenizer=tokenizer2,
            max_seq_length=max_seq_length,
            position_start_id=position_start_id,
        )

        self.vision_processor = Compose(
            [
                Resize(image_size),
                CenterCrop(image_size),
                ToTensor(),
                Normalize([0.5], [0.5]),
            ]
        )
        self.condition_vision_processor = Compose(
            [
                Resize(image_size),
                CenterCrop(image_size),
                ToTensor(),
            ]
        )

        vae_config_dict = json.load(open(vae_config_path))
        vae_scale_factor = 2 ** (len(vae_config_dict.get("block_out_channels", [])) - 1)
        self.vae_image_processor = VaeImageProcessor(
            vae_scale_factor=vae_scale_factor, do_convert_rgb=True
        )
        self.vae_mask_image_processor = VaeImageProcessor(
            vae_scale_factor=vae_scale_factor
        )
        self.vae_condition_image_processor = VaeImageProcessor(
            vae_scale_factor=vae_scale_factor,
            do_convert_rgb=True,
            do_normalize=False,
        )

    def text2image(
        self,
        image: Union[Image.Image, str],
        condition_image: Union[Image.Image, str],
        prompt: str,
        prompt2: Optional[str] = None,
        max_seq_length: Optional[int] = None,
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")
        if isinstance(condition_image, str):
            condition_image = Image.open(condition_image)
        condition_image = condition_image.convert("RGB")

        original_size = image.size
        image = self.vision_resize(image)
        if self.center_crop:
            y1 = max(0, int(round((image.height - self.image_size) / 2.0)))
            x1 = max(0, int(round((image.width - self.image_size) / 2.0)))
            image = self.vision_crop(image)
        else:
            y1, x1, h, w = self.vision_crop.get_params(
                image, (self.image_size, self.image_size)
            )
            image = crop(image, y1, x1, h, w)
        if self.vision_flip:
            x1 = image.width - x1
            image = self.vision_flip(image)
        crop_top_left = (y1, x1)
        pixel_values = self.vision_processor(image)

        add_time_ids = (
            original_size + crop_top_left + [self.image_size, self.image_size]
        )

        condition_pixel_values = self.condition_vision_processor(condition_image)

        prompt2 = prompt2 or prompt
        prompt_outputs = self.text_processor1.classification(
            prompt, max_seq_length=max_seq_length
        )
        prompt2_outputs = self.text_processor2.classification(
            prompt2, max_seq_length=max_seq_length
        )

        return GenericOutputs(
            pixel_values=pixel_values,
            condition_pixel_values=condition_pixel_values,
            input_ids=prompt_outputs.input_ids,
            attention_mask=prompt_outputs.attention_mask,
            input2_ids=prompt2_outputs.input_ids,
            attention2_mask=prompt2_outputs.attention_mask,
            add_time_ids=torch.tensor(add_time_ids),
        )

    def text2image_inputs(
        self,
        condition_image: Union[Image.Image, str],
        prompt: str,
        prompt2: Optional[str] = None,
        negative_prompt: Optional[str] = "",
        negative_prompt2: Optional[str] = None,
        max_seq_length: Optional[int] = None,
    ):
        if isinstance(condition_image, str):
            condition_image = Image.open(condition_image)
        condition_image = condition_image.convert("RGB")

        condition_pixel_values = self.vae_condition_image_processor.preprocess(
            condition_image
        )[0]

        prompt_outputs = self.text_processor1.classification(
            prompt, max_seq_length=max_seq_length
        )
        prompt2_outputs = self.text_processor2.classification(
            prompt2, max_seq_length=max_seq_length
        )
        negative_prompt_outputs = self.text_processor1.classification(
            negative_prompt, max_seq_length=max_seq_length
        )
        negative_prompt2_outputs = self.text_processor2.classification(
            negative_prompt2, max_seq_length=max_seq_length
        )

        return GenericOutputs(
            condition_pixel_values=condition_pixel_values,
            input_ids=prompt_outputs.input_ids,
            attention_mask=prompt_outputs.attention_mask,
            input2_ids=prompt2_outputs.input_ids,
            attention2_mask=prompt2_outputs.attention_mask,
            negative_input_ids=negative_prompt_outputs.input_ids,
            negative_attention_mask=negative_prompt_outputs.attention_mask,
            negative_input2_ids=negative_prompt2_outputs.input_ids,
            negative_attention2_mask=negative_prompt2_outputs.attention_mask,
        )

    def image2image_inputs(
        self,
        prompt: str,
        condition_image: Union[Image.Image, str],
        image: Union[Image.Image, str],
        prompt2: Optional[str] = None,
        negative_prompt: Optional[str] = "",
        negative_prompt2: Optional[str] = None,
        max_seq_length: Optional[int] = None,
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")

        pixel_values = self.vae_image_processor.preprocess(image)[0]

        text_inputs = self.text2image_inputs(
            condition_image=condition_image,
            prompt=prompt,
            prompt2=prompt2,
            negative_prompt=negative_prompt,
            negative_prompt2=negative_prompt2,
            max_seq_length=max_seq_length,
        )
        return GenericOutputs(
            pixel_values=pixel_values,
            **text_inputs,
        )

    def inpainting_inputs(
        self,
        image: Union[Image.Image, str],
        mask_image: Union[Image.Image, str],
        condition_image: Union[Image.Image, str],
        prompt: str,
        prompt2: Optional[str] = None,
        negative_prompt: Optional[str] = "",
        negative_prompt2: Optional[str] = None,
        max_seq_length: Optional[int] = None,
    ):
        if isinstance(image, str):
            image = Image.open(image)
        image = image.convert("RGB")

        if isinstance(mask_image, str):
            mask_image = Image.open(mask_image)
        mask_image = mask_image.convert("L")

        pixel_values = self.vae_image_processor.preprocess(image)[0]
        pixel_masks = self.vae_mask_image_processor.preprocess(mask_image)[0]
        pixel_masks = (pixel_masks + 1) / 2

        text_inputs = self.text2image_inputs(
            condition_image=condition_image,
            prompt=prompt,
            prompt2=prompt2,
            negative_prompt=negative_prompt,
            negative_prompt2=negative_prompt2,
            max_seq_length=max_seq_length,
        )
        return GenericOutputs(
            pixel_values=pixel_values,
            pixel_masks=pixel_masks,
            **text_inputs,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import sys
import logging
import torch
import importlib
import importlib.metadata as importlib_metadata
from unitorch import is_offline_debug_mode

# deepspeed
_deepspeed_available = importlib.util.find_spec("deepspeed") is not None
try:
    _deepspeed_version = importlib_metadata.version("deepspeed")
    logging.debug(f"Successfully imported deepspeed version {_deepspeed_version}")
except importlib_metadata.PackageNotFoundError:
    _deepspeed_available = False


def is_deepspeed_available():
    return _deepspeed_available or is_offline_debug_mode()


# accelerate
_accelerate_available = importlib.util.find_spec("accelerate") is not None
try:
    _accelerate_version = importlib_metadata.version("accelerate")
    logging.debug(f"Successfully imported accelerate version {_accelerate_version}")
except importlib_metadata.PackageNotFoundError:
    _accelerate_available = False


def is_accelerate_available():
    return _accelerate_available or is_offline_debug_mode()


# megatron
_megatron_available = importlib.util.find_spec("megatron") is not None
try:
    _megatron_version = importlib_metadata.version("megatron")
    logging.debug(f"Successfully imported megatron version {_megatron_version}")
except importlib_metadata.PackageNotFoundError:
    _megatron_available = False


def is_megatron_available():
    return _megatron_available or is_offline_debug_mode()


# fastapi
_fastapi_available = importlib.util.find_spec("fastapi") is not None
try:
    _fastapi_version = importlib_metadata.version("fastapi")
    logging.debug(f"Successfully imported fastapi version {_fastapi_version}")
except importlib_metadata.PackageNotFoundError:
    _fastapi_available = False


def is_fastapi_available():
    return _fastapi_available or is_offline_debug_mode()


# diffusers
_diffusers_available = importlib.util.find_spec("diffusers") is not None
try:
    _diffusers_version = importlib_metadata.version("diffusers")
    logging.debug(f"Successfully imported diffusers version {_diffusers_version}")
except importlib_metadata.PackageNotFoundError:
    _diffusers_available = False


def is_diffusers_available():
    return _diffusers_available or is_offline_debug_mode()


# safetensors
_safetensors_available = importlib.util.find_spec("safetensors") is not None
try:
    _safetensors_version = importlib_metadata.version("safetensors")
    logging.debug(f"Successfully imported safetensors version {_safetensors_version}")
except importlib_metadata.PackageNotFoundError:
    _safetensors_available = False


def is_safetensors_available():
    return _safetensors_available or is_offline_debug_mode()


# xformers
_xformers_available = importlib.util.find_spec("xformers") is not None
try:
    _xformers_version = importlib_metadata.version("xformers")
    logging.debug(f"Successfully imported xformers version {_xformers_version}")
except importlib_metadata.PackageNotFoundError:
    _xformers_available = False


def is_xformers_available():
    return _xformers_available or is_offline_debug_mode()


# opencv
_opencv_available = importlib.util.find_spec("cv2") is not None
if _opencv_available:
    logging.debug(f"Successfully imported opencv")


def is_opencv_available():
    return _opencv_available or is_offline_debug_mode()


# torch
_torch_available = importlib.util.find_spec("torch") is not None
try:
    _torch_version = importlib_metadata.version("torch")
    logging.debug(f"Successfully imported torch version {_torch_version}")
except importlib_metadata.PackageNotFoundError:
    _torch_available = False


def is_torch_available():
    return _torch_available or is_offline_debug_mode()


def is_torch2_available():
    return _torch_available and _torch_version >= "2.0.0"


# bitsandbytes
_bitsandbytes_available = importlib.util.find_spec("bitsandbytes") is not None
try:
    _bitsandbytes_version = importlib_metadata.version("bitsandbytes")
    logging.debug(f"Successfully imported bitsandbytes version {_bitsandbytes_version}")
except importlib_metadata.PackageNotFoundError:
    _bitsandbytes_available = False


def is_bitsandbytes_available():
    return _bitsandbytes_available or is_offline_debug_mode()


# auto_gptq
_auto_gptq_available = importlib.util.find_spec("auto_gptq") is not None
try:
    _auto_gptq_version = importlib_metadata.version("auto_gptq")
    logging.debug(f"Successfully imported auto_gptq version {_auto_gptq_version}")
except importlib_metadata.PackageNotFoundError:
    _auto_gptq_available = False


def is_auto_gptq_available():
    return _auto_gptq_available or is_offline_debug_mode()



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.cli.pipelines.controlnet.text2image import (
    ControlNetForText2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.controlnet.image2image import (
    ControlNetForImage2ImageGenerationPipeline,
)
from unitorch.cli.pipelines.controlnet.inpainting import (
    ControlNetForImageInpaintingPipeline,
)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.minigpt4.modeling import (
    MiniGPT4Blip2LlamaModel,
    MiniGPT4Blip2LlamaForGeneration,
)
from unitorch.models.minigpt4.processing import MiniGPT4Blip2LlamaProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import itertools
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.scores import (
    convert_tensor_to_strings,
    remove_strings_ignore_tokens,
)


def _get_ngrams(n, text):
    """Calcualtes n-grams.
    Args:
      n: which n-grams to calculate
      text: An array of tokens
    Returns:
      A set of n-grams
    """
    ngram_set = set()
    text_length = len(text)
    max_index_ngram_start = text_length - n
    for i in range(max_index_ngram_start + 1):
        ngram_set.add(tuple(text[i : i + n]))
    return ngram_set


def _split_into_words(sentences):
    """Splits multiple sentences into words and flattens the result"""
    return list(itertools.chain(*[_.split(" ") for _ in sentences]))


def _get_word_ngrams(n, sentences):
    """Calculates word n-grams for multiple sentences."""
    assert len(sentences) > 0
    assert n > 0

    words = _split_into_words(sentences)
    return _get_ngrams(n, words)


def _len_lcs(x, y):
    """
    Returns the length of the Longest Common Subsequence between sequences x
    and y.
    Source: http://www.algorithmist.com/index.php/Longest_Common_Subsequence
    Args:
      x: sequence of words
      y: sequence of words
    Returns
      integer: Length of LCS between x and y
    """
    table = _lcs(x, y)
    n, m = len(x), len(y)
    return table[n, m]


def _lcs(x, y):
    """
    Computes the length of the longest common subsequence (lcs) between two
    strings. The implementation below uses a DP programming algorithm and runs
    in O(nm) time where n = len(x) and m = len(y).
    Source: http://www.algorithmist.com/index.php/Longest_Common_Subsequence
    Args:
      x: collection of words
      y: collection of words
    Returns:
      Table of dictionary of coord and len lcs
    """
    n, m = len(x), len(y)
    table = dict()
    for i in range(n + 1):
        for j in range(m + 1):
            if i == 0 or j == 0:
                table[i, j] = 0
            elif x[i - 1] == y[j - 1]:
                table[i, j] = table[i - 1, j - 1] + 1
            else:
                table[i, j] = max(table[i - 1, j], table[i, j - 1])
    return table


def _recon_lcs(x, y):
    """
    Returns the Longest Subsequence between x and y.
    Source: http://www.algorithmist.com/index.php/Longest_Common_Subsequence
    Args:
      x: sequence of words
      y: sequence of words
    Returns:
      sequence: LCS of x and y
    """
    i, j = len(x), len(y)
    table = _lcs(x, y)

    def _recon(i, j):
        """private recon calculation"""
        if i == 0 or j == 0:
            return []
        elif x[i - 1] == y[j - 1]:
            return _recon(i - 1, j - 1) + [(x[i - 1], i)]
        elif table[i - 1, j] > table[i, j - 1]:
            return _recon(i - 1, j)
        else:
            return _recon(i, j - 1)

    recon_tuple = tuple(map(lambda x: x[0], _recon(i, j)))
    return recon_tuple


def _multi_rouge_n(sequences, scores_ids, n=2):
    """
    Efficient way to compute highly repetitive scoring
    i.e. sequences are involved multiple time
    Args:
        sequences(list[str]): list of sequences (either hyp or ref)
        scores_ids(list[tuple(int)]): list of pairs (hyp_id, ref_id)
            ie. scores[i] = _rouge_n(scores_ids[i][0],
                                    scores_ids[i][1])
    Returns:
        scores: list of length `len(scores_ids)` containing rouge `n`
                scores as a dict with 'f', 'r', 'p'
    Raises:
        KeyError: if there's a value of i in scores_ids that is not in
                  [0, len(sequences)[
    """
    ngrams = [_get_word_ngrams(n, sequence) for sequence in sequences]
    counts = [len(ngram) for ngram in ngrams]

    scores = []
    for hyp_id, ref_id in scores_ids:
        evaluated_ngrams = ngrams[hyp_id]
        evaluated_count = counts[hyp_id]

        reference_ngrams = ngrams[ref_id]
        reference_count = counts[ref_id]

        overlapping_ngrams = evaluated_ngrams.intersection(reference_ngrams)
        overlapping_count = len(overlapping_ngrams)

        scores += [_f_r_p_rouge_n(evaluated_count, reference_count, overlapping_count)]
    return scores


def _rouge_n(evaluated_sentences, reference_sentences, n=2):
    """
    Computes ROUGE-N of two text collections of sentences.
    Sourece: http://research.microsoft.com/en-us/um/people/cyl/download/
    papers/rouge-working-note-v1.3.1.pdf
    Args:
      evaluated_sentences: The sentences that have been picked by the
                           summarizer
      reference_sentences: The sentences from the referene set
      n: Size of ngram.  Defaults to 2.
    Returns:
      A tuple (f1, precision, recall) for ROUGE-N
    Raises:
      ValueError: raises exception if a param has len <= 0
    """
    if len(evaluated_sentences) <= 0:
        raise ValueError("Hypothesis is empty.")
    if len(reference_sentences) <= 0:
        raise ValueError("Reference is empty.")

    evaluated_ngrams = _get_word_ngrams(n, evaluated_sentences)
    reference_ngrams = _get_word_ngrams(n, reference_sentences)
    reference_count = len(reference_ngrams)
    evaluated_count = len(evaluated_ngrams)

    # Gets the overlapping ngrams between evaluated and reference
    overlapping_ngrams = evaluated_ngrams.intersection(reference_ngrams)
    overlapping_count = len(overlapping_ngrams)

    return _f_r_p_rouge_n(evaluated_count, reference_count, overlapping_count)


def _f_r_p_rouge_n(evaluated_count, reference_count, overlapping_count):
    # Handle edge case. This isn't mathematically correct, but it's good enough
    if evaluated_count == 0:
        precision = 0.0
    else:
        precision = overlapping_count / evaluated_count

    if reference_count == 0:
        recall = 0.0
    else:
        recall = overlapping_count / reference_count

    f1_score = 2.0 * ((precision * recall) / (precision + recall + 1e-8))

    return {"f": f1_score, "p": precision, "r": recall}


def _union_lcs(evaluated_sentences, reference_sentence, prev_union=None):
    """
    Returns LCS_u(r_i, C) which is the LCS score of the union longest common
    subsequence between reference sentence ri and candidate summary C.
    For example:
    if r_i= w1 w2 w3 w4 w5, and C contains two sentences: c1 = w1 w2 w6 w7 w8
    and c2 = w1 w3 w8 w9 w5, then the longest common subsequence of r_i and c1
    is "w1 w2" and the longest common subsequence of r_i and c2 is "w1 w3 w5".
    The union longest common subsequence of r_i, c1, and c2 is "w1 w2 w3 w5"
    and LCS_u(r_i, C) = 4/5.
    Args:
      evaluated_sentences: The sentences that have been picked by the
                           summarizer
      reference_sentence: One of the sentences in the reference summaries
    Returns:
      float: LCS_u(r_i, C)
    ValueError:
      Raises exception if a param has len <= 0
    """
    if prev_union is None:
        prev_union = set()

    if len(evaluated_sentences) <= 0:
        raise ValueError("Collections must contain at least 1 sentence.")

    lcs_union = prev_union
    prev_count = len(prev_union)
    reference_words = _split_into_words([reference_sentence])

    combined_lcs_length = 0
    for eval_s in evaluated_sentences:
        evaluated_words = _split_into_words([eval_s])
        lcs = set(_recon_lcs(reference_words, evaluated_words))
        combined_lcs_length += len(lcs)
        lcs_union = lcs_union.union(lcs)

    new_lcs_count = len(lcs_union) - prev_count
    return new_lcs_count, lcs_union


def _rouge_l_summary_level(evaluated_sentences, reference_sentences):
    """
    Computes ROUGE-L (summary level) of two text collections of sentences.
    http://research.microsoft.com/en-us/um/people/cyl/download/papers/
    rouge-working-note-v1.3.1.pdf
    Calculated according to:
    R_lcs = SUM(1, u)[LCS<union>(r_i,C)]/m
    P_lcs = SUM(1, u)[LCS<union>(r_i,C)]/n
    F_lcs = ((1 + beta^2)*R_lcs*P_lcs) / (R_lcs + (beta^2) * P_lcs)
    where:
    SUM(i,u) = SUM from i through u
    u = number of sentences in reference summary
    C = Candidate summary made up of v sentences
    m = number of words in reference summary
    n = number of words in candidate summary
    Args:
      evaluated_sentences: The sentences that have been picked by the
                           summarizer
      reference_sentence: One of the sentences in the reference summaries
    Returns:
      A float: F_lcs
    Raises:
      ValueError: raises exception if a param has len <= 0
    """
    if len(evaluated_sentences) <= 0 or len(reference_sentences) <= 0:
        raise ValueError("Collections must contain at least 1 sentence.")

    # total number of words in reference sentences
    m = len(set(_split_into_words(reference_sentences)))

    # total number of words in evaluated sentences
    n = len(set(_split_into_words(evaluated_sentences)))

    # print("m,n %d %d" % (m, n))
    union_lcs_sum_across_all_references = 0
    union = set()
    for ref_s in reference_sentences:
        lcs_count, union = _union_lcs(evaluated_sentences, ref_s, prev_union=union)
        union_lcs_sum_across_all_references += lcs_count

    llcs = union_lcs_sum_across_all_references
    r_lcs = llcs / m
    p_lcs = llcs / n
    beta = p_lcs / (r_lcs + 1e-12)
    num = (1 + (beta**2)) * r_lcs * p_lcs
    denom = r_lcs + ((beta**2) * p_lcs)
    f_lcs = num / (denom + 1e-12)
    return {"f": f_lcs, "p": p_lcs, "r": r_lcs}


def rouge1_score(
    y_true: List[Union[str, int, List[Union[str, int]]]],
    y_pred: List[Union[str, int, List[Union[str, int]]]],
    ignore_tokens: Optional[List[Union[str, int]]] = None,
):
    """
    Args:
        y_true: list of lists of int/str tokens of ground truth.
        y_pred: list of lists of int/str tokens of generation results.
        ignore_tokens: the token list to filtration
    """

    if isinstance(y_true, torch.Tensor):
        if y_true.dim() == 2:
            y_true = y_true.unsqueeze(1)
        y_true = convert_tensor_to_strings(y_true)

    if isinstance(y_pred, torch.Tensor) and y_pred.dim() == 2:
        y_pred = convert_tensor_to_strings(y_pred)

    if ignore_tokens is not None:
        ignore_tokens = [str(t) for t in ignore_tokens]

    y_true = remove_strings_ignore_tokens(y_true, ignore_tokens=ignore_tokens)
    y_pred = remove_strings_ignore_tokens(y_pred, ignore_tokens=ignore_tokens)

    y_true = [_y_true[0] for _y_true in y_true]

    num = len(y_pred)
    pre, rec, f1 = 0, 0, 0
    for t, p in zip(y_true, y_pred):
        r = _rouge_n(p, t, n=1)
        pre += r["p"]
        rec += r["r"]
        f1 += r["f"]
    pre, rec, f1 = pre / num, rec / num, f1 / num
    return dict(pre=pre, rec=rec, f1=f1)


def rouge2_score(
    y_true: List[Union[str, int, List[Union[str, int]]]],
    y_pred: List[Union[str, int, List[Union[str, int]]]],
    ignore_tokens: Optional[List[Union[str, int]]] = None,
):
    """
    Args:
        y_true: list of lists of int/str tokens of ground truth.
        y_pred: list of lists of int/str tokens of generation results.
        ignore_tokens: the token list to filtration
    """

    if isinstance(y_true, torch.Tensor):
        if y_true.dim() == 2:
            y_true = y_true.unsqueeze(1)
        y_true = convert_tensor_to_strings(y_true)

    if isinstance(y_pred, torch.Tensor) and y_pred.dim() == 2:
        y_pred = convert_tensor_to_strings(y_pred)

    if ignore_tokens is not None:
        ignore_tokens = [str(t) for t in ignore_tokens]

    y_true = remove_strings_ignore_tokens(y_true, ignore_tokens=ignore_tokens)
    y_pred = remove_strings_ignore_tokens(y_pred, ignore_tokens=ignore_tokens)

    y_true = [_y_true[0] for _y_true in y_true]

    num = len(y_pred)
    pre, rec, f1 = 0, 0, 0
    for t, p in zip(y_true, y_pred):
        r = _rouge_n(p, t, n=2)
        pre += r["p"]
        rec += r["r"]
        f1 += r["f"]
    pre, rec, f1 = pre / num, rec / num, f1 / num
    return dict(pre=pre, rec=rec, f1=f1)


def rougel_score(
    y_true: List[Union[str, int, List[Union[str, int]]]],
    y_pred: List[Union[str, int, List[Union[str, int]]]],
    ignore_tokens: Optional[List[Union[str, int]]] = None,
):
    """
    Args:
        y_true: list of lists of int/str tokens of ground truth.
        y_pred: list of lists of int/str tokens of generation results.
        ignore_tokens: the token list to filtration
    """

    if isinstance(y_true, torch.Tensor):
        if y_true.dim() == 2:
            y_true = y_true.unsqueeze(1)
        y_true = convert_tensor_to_strings(y_true)

    if isinstance(y_pred, torch.Tensor) and y_pred.dim() == 2:
        y_pred = convert_tensor_to_strings(y_pred)

    if ignore_tokens is not None:
        ignore_tokens = [str(t) for t in ignore_tokens]

    y_true = remove_strings_ignore_tokens(y_true, ignore_tokens=ignore_tokens)
    y_pred = remove_strings_ignore_tokens(y_pred, ignore_tokens=ignore_tokens)

    y_true = [_y_true[0] for _y_true in y_true]

    num = len(y_pred)
    pre, rec, f1 = 0, 0, 0
    for t, p in zip(y_true, y_pred):
        r = _rouge_l_summary_level(p, t)
        pre += r["p"]
        rec += r["r"]
        f1 += r["f"]
    pre, rec, f1 = pre / num, rec / num, f1 / num
    return dict(pre=pre, rec=rec, f1=f1)
